/***************************************************************************
                          net2.cpp  -  description
                             -------------------
    begin                : Thu Feb 19 2004
    copyright            : (C) 2004 by Gene Ruebsamen
    email                : gene@localhost.localdomain
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <cmath>
#include <cstdlib>
#include "net2.h"

template class Brain<float>;
template class Brain<double>;
template class Net2<float>;
template class Net2<double>;

template <class T>
Net2<T>::Net2(int in, int hid, int out, int recur) :
input_size(in), hidden_size(hid), output_size(out), recurrent_size(recur),
time(0)
{
	whi = new T[hidden_size*input_size];		// input hidden weights
	whr = new T*[hidden_size*hidden_size];	// recurrent layer
	for (int i=0;i<hidden_size*hidden_size;i++)
		whr[i] = new T[recurrent_size];
	whb = new T[hidden_size];
	woi = new T[output_size*hidden_size];
	wob = new T[output_size];
	// outputs
	ho = new T[hidden_size];			// hidden layer outputs
	oo = new T[output_size];			// output layer outputs
	hiddencopy = new T*[hidden_size];		// for feedback
	for (int i=0;i<hidden_size;i++)
		hiddencopy[i] = new T[recurrent_size];

	vector<T> weights;
	for (int i=0;i<getGeneSize();++i)
		weights.push_back(RandClamped());

	putWeights(weights);
}


template <class T>
Net2<T>::Net2(const Net2<T> &copy) : Brain<T>(copy),
input_size(copy.input_size), hidden_size(copy.hidden_size), output_size(copy.output_size), recurrent_size(copy.recurrent_size),
time(0)
{
	whi = new T[hidden_size*input_size];
	whr = new T*[hidden_size*hidden_size];	// recurrent layer
	for (int i=0;i<hidden_size*hidden_size;i++)
		whr[i] = new T[recurrent_size];
	whb = new T[hidden_size];
	woi = new T[output_size*hidden_size];
	wob = new T[output_size];
	// outputs
	ho = new T[hidden_size];			// hidden layer outputs
	oo = new T[output_size];			// output layer outputs
	hiddencopy = new T* [hidden_size];		// for feedback
	for (int i=0;i<hidden_size;i++)
		hiddencopy[i] = new T[recurrent_size];

	for (int i=0;i<hidden_size;i++)
	{
		whb[i] = copy.whb[i];
		ho[i] = copy.ho[i];
	}

	// recurrent weights (hidden*hidden)
	for (int i=0;i<recurrent_size;i++)
		for (int j=0;j<hidden_size*hidden_size;j++) {
			whr[j][i] = copy.whr[j][i];
		}

	// hidden copy layer (size = hidden)
	for (int i=0;i<recurrent_size;i++)
		for (int j=0;j<hidden_size;j++) {
			hiddencopy[j][i] = copy.hiddencopy[j][i];
		}

	for (int i=0;i<output_size;i++)
	{
		wob[i] = copy.wob[i];
		oo[i] = copy.oo[i];
	}
	for (int i=0;i<(hidden_size*input_size);i++)
		whi[i] = copy.whi[i];
	for (int i=0;i<(output_size*hidden_size);i++)
		woi[i] = copy.woi[i];
}


template <class T>
const Net2<T> &Net2<T>::operator=(const Net2<T> &right)
{
	if (&right != this)	{			// check for self-assignment
		if ((input_size != right.input_size) || (hidden_size != right.hidden_size) || (output_size != right.output_size)
				|| (recurrent_size != right.recurrent_size)) {
			delete [] ho; delete [] oo;
  		delete [] whb; delete [] whi;  //delete [] whr2;
			delete [] wob; delete [] woi; //delete [] wib;
			for (int i=0;i<hidden_size;i++)
				delete [] hiddencopy[i];
			for (int i=0;i<hidden_size*hidden_size;i++)
				delete [] whr[i];

			input_size = right.input_size;
			hidden_size = right.hidden_size;
			output_size = right.output_size;
			whr = new T* [hidden_size*hidden_size];	// recurrant layer
			for (int i=0;i<hidden_size*hidden_size;i++)
				whr[i] = new T[recurrent_size];
			whi = new T[hidden_size*input_size];
			whb = new T[hidden_size];
			woi = new T[output_size*hidden_size];
			wob = new T[output_size];
			// outputs
			ho = new T[hidden_size];			// hidden layer outputs
			oo = new T[output_size];			// output layer outputs
			hiddencopy = new T* [hidden_size];		// for feedback
			for (int i=0;i<hidden_size;i++)
				hiddencopy[i] = new T[recurrent_size];
		}
	}
	// copy items over for net
	for (int i=0;i<hidden_size;i++)
	{
		whb[i] = right.whb[i];
		ho[i] = right.ho[i];
	}

	// copy recurrent weights
	for (int i=0;i<recurrent_size;i++)
		for (int j=0;j<hidden_size*hidden_size;j++) {
			whr[j][i] = right.whr[j][i];
		}

	// copy hidden layers
	for (int i=0;i<recurrent_size;i++)
		for (int j=0;j<hidden_size;j++) {
			hiddencopy[j][i] = right.hiddencopy[j][i];
		}

	for (int i=0;i<output_size;i++)
	{
		wob[i] = right.wob[i];
		oo[i] = right.oo[i];
	}
	for (int i=0;i<(hidden_size*input_size);i++)
		whi[i] = right.whi[i];
	for (int i=0;i<(output_size*hidden_size);i++)
		woi[i] = right.woi[i];
	return *this;
}

template <class T>
void Net2<T>::putWeights(vector<T> &input)
{
  int count = 0;
	int i;
	// setup hidden weight array
	for (i=0;i<(input_size*hidden_size);i++)
		whi[i] = input[count++];
	// setup hidden recurrent weights
	for (int i=0;i<recurrent_size;i++)
		for (int j=0;j<hidden_size*hidden_size;j++)
			whr[j][i] = input[count++];
	// setup hidden bias
	for (i=0;i<hidden_size;i++)
		whb[i] = input[count++];
	// setup output weight array
	for (i=0;i<(hidden_size*output_size);i++)
		woi[i] = input[count++];
	// setup output bias
	for (i=0;i<(output_size);i++)
		wob[i] = input[count++];

	assert (count == input.size());
}

template <class T>
Net2<T>::~Net2()
{
	delete [] ho;  delete [] oo;
	delete [] whb; delete [] whi; //delete [] whr2;
  delete [] wob; delete [] woi;
	for (int i=0;i<hidden_size;i++)
		delete [] hiddencopy[i];
	for (int i=0;i<hidden_size*hidden_size;i++)
		delete [] whr[i];
}

template <class T>
vector<T> Net2<T>::activate(vector<T> &input)
{
	// calculate the output from the hidden layer
	for (int i=0;i<hidden_size;i++)
	{
		T sum=0.0;
		for (int j=0;j<input_size;j++) {
			sum += (input[j]*whi[j+(i*input_size)]);
			//cout << "[" << input[j] << "," << whi[j+(i*input_size)] << "] ";
		}

		if (time > 0) {
			for (int k=0;k<time;k++)         // change k=1 for elman, k=0 for gene's technique
				for (int m=0;m<hidden_size;m++) {
					sum += (hiddencopy[m][k]*whr[i*hidden_size+m][k]);
					//cout << "[" << hiddencopy[m][k] << "," << whr[i*hidden_size+m][k] << "] ";
				}
		}

		sum += whb[i];												// bias term
    //cout << "DONE[OLD]" << endl;
		
		ho[i] = (2.0/(1.0+exp(-2*sum)))-1;         	// use bipolar sigmoid activation function
		hiddencopy[i][0] = ho[i];                		// copy hidden layer  (******* - self modifying on accident? see time[0])
	}

	if (time >= 1) {
		for (int i=recurrent_size-2;i>=0;i--)					// copy hidden layers
			for (int j=0;j<hidden_size;j++)
				hiddencopy[j][i+1] = hiddencopy[j][i];
	}

	vector<T> out;

	// calculate the output from the output layer
	for (int i=0;i<output_size;i++)
	{
		T sum=0.0;
		for (int j=0;j<hidden_size;j++)
		{
			sum += (ho[j]*woi[j+(i*hidden_size)]);
		}
		sum += wob[i];														// bias term
    out.push_back( (2.0/(1.0+exp(-2.0*sum)))-1.0 );
	}     
	if (time < recurrent_size)
		time++;

	return out;																	// return output
}

template <class T>
unsigned int Net2<T>::getGeneSize() const
{
	return ((input_size*hidden_size) + (hidden_size*output_size)
					+ (hidden_size+output_size) + (hidden_size*hidden_size*recurrent_size));
}

template <class T>
vector<T> Net2<T>::getWeights() const
{
	vector<T> weights;

	int i;
	// setup hidden weight array
	for (i=0;i<(input_size*hidden_size);i++)
		weights.push_back(whi[i]);
	// setup hidden recurrent weights
	for (int i=0;i<recurrent_size;i++)
		for (int j=0;j<hidden_size*hidden_size;j++)
			weights.push_back(whr[j][i]);
	// setup hidden bias
	for (i=0;i<hidden_size;i++)
		weights.push_back(whb[i]);
	// setup output weight array
	for (i=0;i<(hidden_size*output_size);i++)
		weights.push_back(woi[i]);
	// setup output bias
	for (i=0;i<(output_size);i++)
		weights.push_back(wob[i]);

	assert(weights.size() == getGeneSize());
	return weights;
}
