/***************************************************************************
                          neuralnet.cpp  -  description
                             -------------------
    begin                : Wed Feb 11 2004
    copyright            : (C) 2004 by Gene Ruebsamen
    email                : gene@localhost.localdomain
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "neuralnet.h"

template class Neuron<float>;
template class Neuron<double>;
template class NeuronLayer<float>;
template class NeuronLayer<double>;
template class MemoryLayer<float>;
template class MemoryLayer<double>;

template class Brain<float>;
template class Brain<double>;
template class NeuralNet<float>;
template class NeuralNet<double>;
   
template <class T>
NeuralNet<T>::NeuralNet(unsigned int numInputs, unsigned int hiddenSize, unsigned int outputSize, unsigned int memoryLayers, unsigned int actFunction) :
m_InputSize(numInputs+(memoryLayers*hiddenSize)),
m_HiddenSize(hiddenSize),
m_OutputSize(outputSize),
m_MemorySize(memoryLayers),
m_ActFunction(actFunction)
{
	// allocate memory
	m_MainLayers = new vector<NeuronLayer<T> >;
  m_MemoryLayers = new vector<MemoryLayer<T> >;
	
	// Hidden Layer
	m_MainLayers->push_back(NeuronLayer<T>(m_HiddenSize, m_InputSize));

	// allocate memory Layer & zero-fill
	for (unsigned i=0;i<m_MemorySize;++i)
	{
		m_MemoryLayers->push_back(MemoryLayer<T>(m_HiddenSize));
		for (unsigned j=0;j<m_HiddenSize;++j)
		{
			(*m_MemoryLayers)[i].m_vecNeurons.push_back(0);
		}
	}

	// Output Layer
	m_MainLayers->push_back(NeuronLayer<T>(m_OutputSize, m_HiddenSize));
}

template <class T>
NeuralNet<T>::NeuralNet(const NeuralNet<T> &copy) : Brain<T>(copy)
{
	m_ActFunction = copy.m_ActFunction;
	m_HiddenSize = copy.m_HiddenSize;
	m_InputSize = copy.m_InputSize;
	m_MemorySize = copy.m_MemorySize;
	m_OutputSize = copy.m_OutputSize;
	m_MainLayers = new vector<NeuronLayer<T> >;
	m_MemoryLayers = new vector<MemoryLayer<T> >;

	*m_MainLayers = (*(copy.m_MainLayers));
	*m_MemoryLayers = (*(copy.m_MemoryLayers));
}

template <class T>
NeuralNet<T>::~NeuralNet()
{
	delete m_MainLayers;
	delete m_MemoryLayers;
}

template <class T>
bool NeuralNet<T>::operator==(const NeuralNet<T> &right) const
{
	if (m_ActFunction != right.m_ActFunction)
		return false;
	if (m_HiddenSize != right.m_HiddenSize)
		return false;
	if (m_InputSize != right.m_InputSize)
		return false;
	if (m_MemorySize != right.m_MemorySize)
		return false;
	if (m_OutputSize != right.m_OutputSize)
		return false;
	if (*m_MemoryLayers != *(right.m_MemoryLayers))
		return false;
	if (*m_MainLayers != *(right.m_MainLayers))
		return false;

	return true;
}

template <class T>
const NeuralNet<T> &NeuralNet<T>::operator=(const NeuralNet<T> &right)
{
	if (this != &right)
	{
		m_ActFunction = right.m_ActFunction;
		m_HiddenSize = right.m_HiddenSize;
		m_InputSize = right.m_InputSize;
		m_MemorySize = right.m_MemorySize;
		m_OutputSize = right.m_OutputSize;
		
		*m_MainLayers = (*(right.m_MainLayers));
		*m_MemoryLayers = (*(right.m_MemoryLayers));
	}
	
	return *this;
}

template <class T>
T &NeuralNet<T>::operator[](unsigned int index)
{
	unsigned int count = 0;

	// for the main layers
	for (int i=0;i<2;++i)
	{
		// for each neuron
		for (int j=0;j<(*m_MainLayers)[i].m_NumNeurons;++j)
		{
			// for each weight
			for (int k=0;k<(*m_MainLayers)[i].m_vecNeurons[j].m_NumInputs;++k)
			{
				if (index == count)
					return (*m_MainLayers)[i].m_vecNeurons[j].m_vecWeights[k];
				count++;
			}
		}
	}

	exit(1);
}

template <class T>
vector<T> NeuralNet<T>::getWeights() const
{
	vector<T> weights;

	// for the main layers
	for (int i=0;i<2;++i)
	{
		// for each neuron
		for (int j=0;j<(*m_MainLayers)[i].m_NumNeurons;++j)
		{
			// for each weight
			for (int k=0;k<(*m_MainLayers)[i].m_vecNeurons[j].m_NumInputs;++k)
			{
				weights.push_back((*m_MainLayers)[i].m_vecNeurons[j].m_vecWeights[k]);
			}
		}
	}
	
	return weights;
}

template <class T>
unsigned int NeuralNet<T>::getGeneSize() const
{
	unsigned int count = 0;
	// for the main layers
	for (int i=0;i<2;++i)
	{
		// for each neuron
		for (int j=0;j<(*m_MainLayers)[i].m_NumNeurons;++j)
		{
			// for each weight
			for (int k=0;k<(*m_MainLayers)[i].m_vecNeurons[j].m_NumInputs;++k)
			{
				count++;
			}
		}
	}
	
	return count;
}

template <class T>
void NeuralNet<T>::putWeights(vector<T> &input)
{
	int counter=0;

	// for each main layer
	for (int i=0;i<2;++i)
	{
		// for each neuron
		for (int j=0;j<(*m_MainLayers)[i].m_NumNeurons;++j)
		{
			// for each weight
			for (int k=0;k<(*m_MainLayers)[i].m_vecNeurons[j].m_NumInputs;++k)
			{
				(*m_MainLayers)[i].m_vecNeurons[j].m_vecWeights[k] = input[counter++];
			}
		}
	}

}

/*
template <class T>
vector<T> NeuralNet<T>::activate(vector<T> &input)
{
	// storage for the output
	vector<T> output;

	int counter = 0;
	
	// verify the correct number of inputs
	if (input.size()+m_MemorySize*m_HiddenSize != static_cast<unsigned int>(m_InputSize))
	{
		return output;
	}

	// for the main layers
	for (int i=0;i<2;++i)
	{
		counter = 0;
		
		if (i>0)
		{
			input = output;
			// add hidden layer to memory
			addMemory(output);
		}
		else
		{
			// handle the memory inputs
			for (unsigned j=0;j<m_MemorySize;++j)
			{
				for (unsigned k=0;k<m_HiddenSize;++k)
					input.push_back((*m_MemoryLayers)[j].m_vecNeurons[k]);
			}
		}

		output.clear();
		for (int j=0;j<(*m_MainLayers)[i].m_NumNeurons;++j)
		{
			T netinput = 0;
			int NumInputs = (*m_MainLayers)[i].m_vecNeurons[j].m_NumInputs;

			// for each weight
			for (int k=0;k<NumInputs - 1; ++k)
			{
				netinput += (*m_MainLayers)[i].m_vecNeurons[j].m_vecWeights[k] * input[counter++];
			}

			// add the bias
			netinput += (*m_MainLayers)[i].m_vecNeurons[j].m_vecWeights[NumInputs-1] * BIAS;

			switch(m_ActFunction)
			{
				case 0:
					output.push_back(actSigmoid(netinput));
					break;
				case 1:
					output.push_back(actBipolarSigmoid(netinput));
					break;
				case 2:
					output.push_back(actSine(netinput));
					break;
				default:
					break;
			}
			counter = 0;
		}
	}

	return output;
}
*/

template <class T>
vector<T> NeuralNet<T>::activate(vector<T> &input)
{
	// storage for the output
	vector<T> output;

	int counter = 0;

	// verify the correct number of inputs
	if (input.size()+m_MemorySize*m_HiddenSize != static_cast<unsigned int>(m_InputSize))
	{
		return output;
	}

	// for the main layers
	for (int i=0;i<2;++i)
	{
		counter = 0;

		if (i>0)
		{
			input = output;
			// add hidden layer to memory
			addMemory(output);
		}
		else
		{
			// handle the memory inputs
			for (unsigned j=0;j<m_MemorySize;++j)
			{
				for (unsigned k=0;k<m_HiddenSize;++k)
					input.push_back((*m_MemoryLayers)[j].m_vecNeurons[k]);
			}
		}

		output.clear();
		for (int j=0;j<(*m_MainLayers)[i].m_NumNeurons;++j)
		{
			T netinput = 0;
			int NumInputs = (*m_MainLayers)[i].m_vecNeurons[j].m_NumInputs;

			// for each weight
			for (int k=0;k<NumInputs - 1; ++k)
			{
				netinput += (*m_MainLayers)[i].m_vecNeurons[j].m_vecWeights[k] * input[counter++];
			}

			// add the bias
			netinput += (*m_MainLayers)[i].m_vecNeurons[j].m_vecWeights[NumInputs-1] * BIAS;

			T result;
			switch(m_ActFunction)
			{
				case 0:
					//output.push_back(actSigmoid(netinput));
					result = actSigmoid(netinput);
					break;
				case 1:
					//output.push_back(actBipolarSigmoid(netinput));
					result = actBipolarSigmoid(netinput);
					break;
				case 2:
					//output.push_back(actSine(netinput));
					result = actSine(netinput);
					break;
				default:
					break;
			}
			output.push_back(result);
			
			// EXPERIMENTAL
			// self modifying inputs
			if (i==0) {
				input[counter-(NumInputs-1)+j+(input.size()-(m_MemorySize*m_HiddenSize))]=result;
			}
			// END EXPERIMENTAL

			counter = 0;
		}
	}

	return output;
}


template <class T>
void NeuralNet<T>::addMemory(vector<T> &input)
{
	// move all memory down one layer
	for (unsigned i=m_MemorySize-1;i>0;--i)
	{
		(*m_MemoryLayers)[i] = (*m_MemoryLayers)[i-1];
	}

	// add inputs
	for (unsigned i=0;i<m_HiddenSize;++i)
	{
		(*m_MemoryLayers)[0].m_vecNeurons[i] = input[i];
	}	
}

template <class T>
void NeuralNet<T>::clearMemory()
{
	// clear Memory
	for (unsigned i=0;i<m_MemorySize;++i)
	{
		for (unsigned j=0;j<m_HiddenSize;++j)
		{
			(*m_MemoryLayers)[i].m_vecNeurons[j] = 0;
		}
	}	
}

template <class T>
T NeuralNet<T>::actSigmoid(T input) const
{
	T result = 1.0/(1.0 + exp(-input));
	// output is [0, 1.0]
	return (result);
}

template <class T>
T NeuralNet<T>::actBipolarSigmoid(T input) const
{
	T result = (2.0/(1.0 + exp(-2.0*input)))-1.0;
	// output is [-1.0, 1.0]
	return (result);
}

template <class T>
T NeuralNet<T>::actSine(T input) const
{
	return sin(input);
}
