/***************************************************************************
                          neuralnet.h  -  description
                             -------------------
    begin                : Tue Feb 10 2004
    copyright            : (C) 2004 by Gene Ruebsamen
    email                : gene@localhost.localdomain
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef NEURALNET_H
#define NEURALNET_H

#include <vector>
#include <cmath>
#include <iostream>
#include "../constants.h"
#include "neuron.h"
#include "brain.h"

using namespace std;
using std::vector;

template <class T>
class NeuralNet : public Brain<T>
{
public:
	NeuralNet(unsigned int inputSize, unsigned int hiddenSize, unsigned int outputSize, unsigned int numMemoryLayers, unsigned int actFunction=0);
	NeuralNet(const NeuralNet &);
	virtual ~NeuralNet();
	vector<T> getWeights() const;
	unsigned int getGeneSize() const;
	unsigned int getInputSize() const
		{ return m_InputSize; }
	unsigned int getHiddenSize() const
		{ return m_HiddenSize; }
	unsigned int getOutputSize() const
		{ return m_OutputSize; }
	unsigned int getMemoryLayerSize() const
		{ return m_MemorySize; }
	unsigned int getActivationType() const
		{ return m_ActFunction; }
	void putWeights(vector<T> &);
	vector<T> activate(vector<T> &input);
	void clearMemory();

	bool operator==(const NeuralNet<T> &) const;
	bool operator!=(const NeuralNet<T> &right) const
		{ return !(*this == right); }
	inline T &operator[](unsigned int);
	const NeuralNet<T> &operator=(const NeuralNet<T> &right);

private:
	// activation functions
	inline T actSine(T) const;
	inline T actSigmoid(T) const;
	inline T actBipolarSigmoid(T) const;

	void addMemory(vector<T> &);
	unsigned int m_InputSize;
	unsigned int m_HiddenSize;
	unsigned int m_OutputSize;
	unsigned int m_MemorySize;
	unsigned int m_ActFunction;
	vector<NeuronLayer<T> > *m_MainLayers;
	vector<MemoryLayer<T> > *m_MemoryLayers;
};

#endif
