/***************************************************************************
                          easyhuman.cpp  -  description
                             -------------------
    begin                : Wed Feb 20 2002
    copyright            : (C) 2002 by Gene Ruebsamen
    email                : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "easyhuman.h"      // import constants for bodies

EasyHuman::EasyHuman(dWorldID world, dSpaceID space, dReal x, dReal y, dReal z) :
	Creature(world,space,x,y,z), numBody(6), numGeom(3), numJoint(5)
{
	// setup
	HEAD = 0.12;         // x y z
	TORSO[0] = 0.4; TORSO[1] = 0.25; TORSO[2] = 0.45;
	LEG1[0] = 0.15; LEG1[1] = 0.15; LEG1[2] = 0.42;
	FOOT[0] = 0.20; FOOT[1] = 0.40; FOOT[2] = 0.05;
	GAP = 0.01;

	l_foot_sens = false;					// init foot sensors
	r_foot_sens = false;
	body_sens = false;						// init body sensor
	fallen = false;
	body = new dBodyID[numBody];	// allocate space for Bodies
	geom = new dGeomID[numGeom];	// allocate space for Geometries	
	joint = new dJointID[numJoint];	// allocate space for Joints
	
	// create new Joint Group
	jGroup = dJointGroupCreate(0);
//	geomGroup = dCreateGeomGroup(space);  slow

	// constants (Mass)
	const dReal MHEAD=0.03;
	const dReal MTORSO=0.6;
	const dReal MULEG=0.5;
	const dReal MFOOT=0.09;

	dMass m;

	// create bodies
	for (int i=0;i<numBody;i++)
		body[i]=dBodyCreate(world);

	dBodySetPosition (body[0],0+x,0+y,z+TORSO[2]/2+HEAD);	// head
	dBodySetPosition (body[1],0+x,0+y,z);	// torso
	dBodySetPosition (body[2],TORSO[0]/3+x,0+y,z-(TORSO[2]/2+LEG1[2]/2+GAP));	//right upper leg
	dBodySetPosition (body[3],-TORSO[0]/3+x,0+y,z-(TORSO[2]/2+LEG1[2]/2+GAP));	// left upper leg
	dBodySetPosition (body[4],TORSO[0]/3+GAP/2+x,-FOOT[0]/4+y,z-TORSO[2]/2-LEG1[2]-8*GAP);	//right foot
	dBodySetPosition (body[5],-TORSO[0]/3-GAP/2+x,-FOOT[0]/4+y,z-TORSO[2]/2-LEG1[2]-8*GAP);	//left foot

	// head
	dMassSetSphere(&m,1,HEAD);
	dMassAdjust(&m,MHEAD);
	dBodySetMass(body[0], &m);

	// torso
	dMassSetBox (&m,1,TORSO[0],TORSO[1],TORSO[2]);	// torso
	dMassAdjust (&m,MTORSO);	// torso weight
	dBodySetMass (body[1],&m);
	geom[0] = dCreateBox (space,TORSO[0],TORSO[1],TORSO[2]);
	dGeomSetBody (geom[0],body[1]);

	// upper legs
	dMassSetBox (&m,1,LEG1[0],LEG1[1],LEG1[2]);	// legs
	dMassAdjust (&m,MULEG);	// upper leg weight
		dBodySetMass (body[2],&m);
		dBodySetMass (body[3],&m);

	// feet
	dMassSetBox (&m,1,FOOT[0],FOOT[1],FOOT[2]); 	// feet
	dMassAdjust (&m,MFOOT);	// feet weight
		dBodySetMass (body[4],&m);
		dBodySetMass (body[5],&m);
		geom[1] = dCreateBox (space,FOOT[0],FOOT[1],FOOT[2]);  // 0
		dGeomSetBody (geom[1],body[4]);
		geom[2] = dCreateBox (space,FOOT[0],FOOT[1],FOOT[2]);  // 0
		dGeomSetBody (geom[2],body[5]);

/*	// add geometries to GeomGroup
	for (int i=0;i<numGeom;i++)
		dGeomGroupAdd(geomGroup,geom[i]); */


	/********************************** CREATE JOINTS **********************************/
	// head joint
	joint[0] = dJointCreateFixed(world,jGroup);		// add to joint group jGroup
	dJointAttach(joint[0],body[1],body[0]);
	dJointSetFixed(joint[0]);

	// right Leg Socket
	joint[1] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[1],body[1],body[2]);
	dJointSetHingeAnchor(joint[1],TORSO[0]/3+x,0+y,z-(TORSO[2]/2)-GAP/2);
	dJointSetHingeAxis(joint[1],1,0,0);

	// left Leg Socket
	joint[2] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[2],body[1],body[3]);
	dJointSetHingeAnchor(joint[2],-TORSO[0]/3+x,0+y,z-(TORSO[2]/2)-GAP/2);
	dJointSetHingeAxis(joint[2],1,0,0);

	for (int i=1;i<=2;i++)
	{
		dJointSetHingeParam(joint[i],dParamLoStop,-1.2);
		dJointSetHingeParam(joint[i],dParamHiStop,1.5);
		dJointSetHingeParam(joint[i],dParamStopERP,0.2);
		dJointSetHingeParam(joint[i],dParamStopCFM,0.05);
		dJointSetHingeParam(joint[i],dParamBounce,0.1);
		dJointSetHingeParam(joint[i],dParamFudgeFactor,0.5);
		dJointSetHingeParam(joint[i],dParamVel,0);
		dJointSetHingeParam(joint[i],dParamFMax,0.5);
	}
	// right foot
	joint[3] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[3],body[2],body[4]);
	dJointSetHingeAnchor(joint[3],TORSO[0]/3+GAP/2+x,0.06+y,z-TORSO[2]/2-LEG1[2]-(8*GAP/2));
	dJointSetHingeAxis(joint[3],1,0,0);
	dJointSetHingeParam(joint[3],dParamLoStop,-0.8);
	dJointSetHingeParam(joint[3],dParamHiStop,0.3);
	dJointSetHingeParam(joint[3],dParamStopERP,0.2);
	dJointSetHingeParam(joint[3],dParamStopCFM,0.09);
	dJointSetHingeParam(joint[3],dParamBounce,0.3);
	dJointSetHingeParam(joint[3],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[3],dParamVel,0);
	dJointSetHingeParam(joint[3],dParamFMax,0.5);
	// left foot
	joint[4] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[4],body[3],body[5]);
	dJointSetHingeAnchor(joint[4],-TORSO[0]/3-GAP/2+x,0.06+y,z-TORSO[2]/2-LEG1[2]-(8*GAP/2));
	dJointSetHingeAxis(joint[4],1,0,0);
	dJointSetHingeParam(joint[4],dParamLoStop,-0.8);
	dJointSetHingeParam(joint[4],dParamHiStop,0.3);
	dJointSetHingeParam(joint[4],dParamStopERP,0.2);
	dJointSetHingeParam(joint[4],dParamStopCFM,0.09);
	dJointSetHingeParam(joint[4],dParamBounce,0.3);
	dJointSetHingeParam(joint[4],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[4],dParamVel,0);
	dJointSetHingeParam(joint[4],dParamFMax,0.5);
}

EasyHuman::~EasyHuman()
{
	// destroy joint group
	dJointGroupDestroy(jGroup);

	// destroy Geometries
	for (int i=0;i<numGeom;i++) {
 //  	dGeomGroupRemove(geomGroup,geom[i]);
		dGeomDestroy(geom[i]);
	}
	// destroy bodies
	for (int i=0;i<numBody;i++)
		dBodyDestroy(body[i]);

	// deallocate space
	delete [] body;
	delete [] geom;
	delete [] joint;
}

void EasyHuman::draw(bool selected) const
{
	dsSetColor (0,1,1);
	dsSetTexture (DS_WOOD); // DS_WOOD
	// head
	dsDrawSphere(dBodyGetPosition(body[0]),dBodyGetRotation(body[0]),HEAD);
	if (selected)
		dsSetColor (1,0,0);
	else
		dsSetColor (1,0,1);
	// torso
	dsDrawBox(dBodyGetPosition(body[1]),dBodyGetRotation(body[1]),TORSO);
	dsSetColor (1,1,0);
	// legs
	dsDrawBox(dBodyGetPosition(body[2]),dBodyGetRotation(body[2]),LEG1);
	dsDrawBox(dBodyGetPosition(body[3]),dBodyGetRotation(body[3]),LEG1);
	dsSetColor (0.5,0,1);
	// feet
	dsDrawBox(dBodyGetPosition(body[4]),dBodyGetRotation(body[4]),FOOT);
	dsDrawBox(dBodyGetPosition(body[5]),dBodyGetRotation(body[5]),FOOT);
}

void EasyHuman::shutoff(void)
{
	// right Leg Socket
	dJointSetHingeParam(joint[1],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[1],dParamFMax, (dReal)0);
	// left leg socket
	dJointSetHingeParam(joint[2],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[2],dParamFMax,(dReal)0);
	// right foot
	dJointSetHingeParam(joint[3],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[3],dParamFMax,(dReal)0);
	// left foot
	dJointSetHingeParam(joint[4],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[4],dParamFMax,(dReal)0);	
}

void EasyHuman::setMuscles(const float * const inputs)
{
	// right Leg Socket
	dJointSetHingeParam(joint[1],dParamVel,(dReal)inputs[0]);
	dJointSetHingeParam(joint[1],dParamFMax,(dReal)3.95);
	// left leg socket
	dJointSetHingeParam(joint[2],dParamVel,(dReal)inputs[1]);
	dJointSetHingeParam(joint[2],dParamFMax,(dReal)3.95);
	// right foot
	dJointSetHingeParam(joint[3],dParamVel,(dReal)inputs[2]);
	dJointSetHingeParam(joint[3],dParamFMax,(dReal)1.0);
	// left foot
	dJointSetHingeParam(joint[4],dParamVel,(dReal)inputs[3]);
	dJointSetHingeParam(joint[4],dParamFMax,(dReal)1.0);	
}

void EasyHuman::getInputs(float * const sensors) const
{
	// position inputs (position & rotation of head)
	const dReal * rot = dBodyGetRotation(body[0]);
	const dReal * rot2 = dBodyGetAngularVel(body[0]);
//	const dReal * rot3 = dBodyGetAngularVel(body[5]);
	
//	sensors[0] = (float)pos[2];		// z (height) of head
	sensors[0] = (float)rot[0];
	sensors[1] = (float)rot[1];
	sensors[2] = (float)rot[2];
	sensors[3] = (float)rot2[0];
	sensors[4] = (float)rot2[1];
  sensors[5] = (float)rot2[2];
	// get right & left leg socket angles (3)
	sensors[6] = (float)dJointGetHingeAngleRate(joint[1]);
	sensors[7] = (float)dJointGetHingeAngleRate(joint[2]);
	// get right foot angle
	sensors[8] = (float)dJointGetHingeAngleRate(joint[3]);
	// get left foot angle
	sensors[9] = (float)dJointGetHingeAngleRate(joint[4]);
	if (r_foot_sens) 	sensors[10] = 1.0;
		else	sensors[10] = -1.0;
	if (l_foot_sens)  sensors[11] = 1.0;
		else	sensors[11] = -1.0;
}

void EasyHuman::footContact(dGeomID o1, dGeomID o2, dGeomID ground)
{
	if (((o1==geom[1]) && (o2==ground)) || ((o2==geom[1])&&(o1==ground)))
		r_foot_sens=true;
	else
		r_foot_sens=false;
	if (((o1==geom[2]) && (o2==ground)) || ((o2==geom[2])&&(o1==ground)))
		l_foot_sens=true;
	else
		l_foot_sens=false;
}

void EasyHuman::bodyContact(dGeomID o1, dGeomID o2, dGeomID ground)
{
	if (((o1==geom[0])&&(o2==ground)) || ((o2==geom[0])&&(o1==ground)))
	{
		body_sens=true;
		fallen=true;
	}
	else
		body_sens=false;
}

const dReal *EasyHuman::getHeadPos(void) const
{
	return dBodyGetPosition(body[0]);
}

const dReal *EasyHuman::getRFootPos(void) const
{
	return dBodyGetPosition(body[4]);
}

const dReal *EasyHuman::getLFootPos(void) const
{
	return dBodyGetPosition(body[5]);
}

bool EasyHuman::hasFallen(void) const
{
	return fallen;
}

float EasyHuman::getFitness(void) const
{
	const dReal * pos = getHeadPos();
	float fitness;

	// fitness function
	if (fabs(pos[1]) > 10 || (fabs(pos[2]) > 5) || (fabs(pos[0]-getStartX()) > 5))
		fitness = 0;
	else if (hasFallen())
		fitness = -pos[1]+(pos[2]*8);  	// if fallen down
	else
		fitness = -(pos[1]*2)+(pos[2]*8); // if still standing

	return fitness;
}

void EasyHuman::randomForce(void) const
{
    dReal f[3];
    f[0] = 22.5 * (-0.5+1.0/RAND_MAX * rand());
    f[1] = 42.5 * (-0.5+1.0/RAND_MAX * rand());
    f[2] = .5 * (-0.5+1.0/RAND_MAX * rand());
    dBodyAddForce(body[0], f[0], f[1], f[2]);
}
void EasyHuman::randomTorque(void) const
{
    dReal f[3];
    f[0] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
    f[1] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
    f[2] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
	  dBodyAddTorque(body[0], f[0], f[1], f[2]);
}