/***************************************************************************
                          genetic.cpp  -  description
                             -------------------
    begin                : Fri Feb 13 2004
    copyright            : (C) 2004 by Gene Ruebsamen
    email                : gene@localhost.localdomain
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "genetic.h"

Genetic::Genetic() :
m_NumElite(1),
m_AvgFitness(0),
m_MaxFitness(0),
m_TotalFitness(0)
{
}

Genetic::~Genetic()
{
}

void Genetic::setNumElite(unsigned int num)
{
	m_NumElite = num;
}

void Genetic::printStats() const
{
	cout << "Gen: " << m_Generation << "\tAvg: " << m_AvgFitness << "\tMax:  " << m_MaxFitness << endl;
}

//--------
// Update the Genetic Algorithm
//--------
vector<Organism<float> *> Genetic::epoch(vector<Organism<float> *> &population)
{
  int max_index = 0;
	m_Generation++;
	m_TotalFitness = 0;
	m_AvgFitness = 0;
	m_MaxFitness = 0;
	// calculate statistics
  for (int i=0;i<population.size();i++)
	{
		double temp_fitness = population[i]->getFitness();
		m_TotalFitness += temp_fitness;
		if (temp_fitness > m_MaxFitness) {
			max_index = i;
			m_MaxFitness = temp_fitness;
		}
	}

	m_AvgFitness = m_TotalFitness / static_cast<double>(population.size());

	// sort population vector
	if (m_NumElite > 0)
		sort(population.begin(),population.end(), std::greater< Organism<float> * >());
	
	// linear fitness scaling routines
/*		const double fmultiple=2.0;
		double delta,a,b, sumfitness, fitness;
		sumfitness=0.0;
		if (min > ((fmultiple*avg - max) / (fmultiple - 1.0))) {
			delta = max - avg;
			a = (fmultiple - 1.0) * avg/delta;
			b = avg * (max - fmultiple*avg) / delta;
		} else {
			delta = avg - min;
			a = avg / delta;
			b = -min * avg / delta;
		}
		for (int j=0;j<POPSIZE;j++)
		{
			fitness = a * fitarray[j] + b;
			fitarray[j]=fitness;
			sumfitness += fitness;
		}	      */

	// create space for new population
	vector<Organism<float> *> offspring;
	offspring.reserve(population.size());

	// roulette wheel selection
	for (int i=0;i<population.size()-m_NumElite;i++)	
	{
		// pick the lucky couple
		Organism<float> *father = roulette_selection(population);
    Organism<float> *mother = roulette_selection(population);
    // mate & produce offspring
		offspring.push_back(*father + *mother);
	}

	// Elitism
	if (m_NumElite > 0)
	{
		vector<Organism<float> *>::iterator iter_start = population.begin();
		vector<Organism<float> *>::iterator iter_end;
		int index=0;
		for (iter_end=population.begin();index<m_NumElite;iter_end++,index++)
		{
			offspring.push_back(*iter_end);
			(*iter_end)->select(true);
		}
		// remove the elite individuals from the original population
		population.erase(iter_start,iter_end);
	}
	
	// recalculate the starting positions of the new population
	for (int i=0;i<offspring.size();++i)
	{
		dVector3 start = {(i*SPACING)-POPSIZE/2,0,STARTZ};
		offspring[i]->setStartPosition(start);
	}
	return offspring;
}


// Roulette Wheel selection on the population of Organisms
Organism<float> *Genetic::roulette_selection(const vector<Organism<float> *> &population)
{
	double roulette, sum=0;
	int count=0;
	roulette = (((float)(rand() %1000))/1000.0)*m_TotalFitness;
	while (sum < roulette)
		sum += population[count++]->getFitness();
	if (count > 0)
		count--;

  assert(count < population.size());
	return population[count];
}
