/***************************************************************************
                          Hopper.cpp  -  description
                             -------------------
    begin                : Thu Feb 21 2002
    copyright            : (C) 2002 by Gene Ruebsamen
    email                :
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "hopper.h"      // import constants for bodies
#include <iostream>

using std::cout;
using std::endl;

Hopper::Hopper(dWorldID world, dSpaceID space, dReal x, dReal y, dReal z) :
	Creature(world,space,x,y,z), numBody(5), numGeom(5), numJoint(4), fitcount(0)		// Build Creature
{
	// Body Setup
	TORSO[0] = 0.3; TORSO[1] = 0.3; TORSO[2] = 0.2;
	LEG1[0] = 0.07; LEG1[1] = 0.07; LEG1[2] = 0.4;
	LEG2[0] = 0.07; LEG2[1] = 0.05; LEG2[2] = 0.3;
	GAP = 0.01;

	r_foot_sens = false;					// init foot sensors
  l_foot_sens = false;
	fallen = false;
	body = new dBodyID[numBody];	// allocate space for Bodies
	geom = new dGeomID[numGeom];	// allocate space for Geometries	
	joint = new dJointID[numJoint];	// allocate space for Joints
	
	// create new Joint Group
	jGroup = dJointGroupCreate(0);
//	geomGroup = dCreateGeomGroup(space);

	// constants (Mass)
	const dReal MTORSO=4;
	const dReal MULEG=0.2;
	const dReal MMLEG=.1;

	dMass m;

	// create bodies
	for (int i=0;i<numBody;i++)
		body[i]=dBodyCreate(world);

	dBodySetPosition (body[0],0+x,0+y,z);	// torso
	dBodySetPosition (body[1],x-TORSO[0]/4,y,z-LEG1[2]/5-GAP);	//right upper leg
	dBodySetPosition (body[2],x+TORSO[0]/4,y,z-LEG1[2]/5-GAP);	//left upper leg
	dBodySetPosition (body[3],x-TORSO[0]/4,y,z-LEG1[2]-LEG2[2]/2-GAP);	//right foot
	dBodySetPosition (body[4],x+TORSO[0]/4,y,z-LEG1[2]-LEG2[2]/2-GAP);	//left foot

	// torso
	dMassSetBox (&m,1,TORSO[0],TORSO[1],TORSO[2]);	// torso
	dMassAdjust (&m,MTORSO);	// torso weight
	dBodySetMass (body[0],&m);
	geom[0] = dCreateBox (space,TORSO[0],TORSO[1],TORSO[2]);
	dGeomSetBody (geom[0],body[0]);

	// upper legs
	dMassSetBox (&m,1,LEG1[0],LEG1[1],LEG1[2]);	// legs
	dMassAdjust (&m,MULEG);	// upper leg weight
	dBodySetMass (body[1],&m);
	geom[1] = dCreateBox (space,LEG1[0],LEG1[1],LEG1[2]);
	dGeomSetBody (geom[1],body[1]);
	dBodySetMass (body[2],&m);
	geom[2] = dCreateBox (space,LEG1[0],LEG1[1],LEG1[2]);
	dGeomSetBody (geom[2],body[2]);

	// lower legs
	dMassSetBox (&m,1,LEG2[0],LEG2[1],LEG2[2]);	// legs
	dMassAdjust (&m,MMLEG);	// middle leg weight
	dBodySetMass (body[3],&m);
	geom[3] = dCreateBox (space,LEG1[0],LEG1[1],LEG1[2]);
	dGeomSetBody (geom[3],body[3]);
	dBodySetMass (body[4],&m);
	geom[4] = dCreateBox (space,LEG1[0],LEG1[1],LEG1[2]);
	dGeomSetBody (geom[4],body[4]);

/*or (int i=0;i<numGeom;i++)
		dGeomGroupAdd(geomGroup,geom[i]); */

	/********************************** CREATE JOINTS **********************************/
	// right Leg Socket
	// right Leg Socket
/*	joint[0] = dJointCreateBall(world,jGroup);
	dJointAttach(joint[0],body[0],body[1]);
	dJointSetBallAnchor(joint[0],x,y,z);
	joint[2] = dJointCreateAMotor(world,jGroup);
	dJointAttach(joint[2],body[0],body[1]);
		dJointSetAMotorMode(joint[2],dAMotorEuler);
		dJointSetAMotorAxis(joint[2],0,1,0,0,1);
		dJointSetAMotorAxis(joint[2],2,2,0,1,0);
		dJointSetAMotorParam(joint[2],dParamLoStop,-0.2);  // twist
		dJointSetAMotorParam(joint[2],dParamHiStop,0.2);
		dJointSetAMotorParam(joint[2],dParamVel, 0);
		dJointSetAMotorParam(joint[2],dParamFMax,25);
		dJointSetAMotorParam(joint[2],dParamLoStop2,-0.5);  //back
		dJointSetAMotorParam(joint[2],dParamHiStop2,0.5);	// front
		dJointSetAMotorParam(joint[2],dParamVel2,0);
		dJointSetAMotorParam(joint[2],dParamFMax2,0.5);
		dJointSetAMotorParam(joint[2],dParamLoStop3,-0.5);  //back
		dJointSetAMotorParam(joint[2],dParamHiStop3,0.5);	// front
		dJointSetAMotorParam(joint[2],dParamVel3,0);
		dJointSetAMotorParam(joint[2],dParamFMax3,0.5);
		dJointSetAMotorParam(joint[2],dParamStopERP,0.3);
		dJointSetAMotorParam(joint[2],dParamStopCFM,0.05);
		dJointSetAMotorParam(joint[2],dParamStopERP2,0.3);
		dJointSetAMotorParam(joint[2],dParamStopCFM2,0.05);
		dJointSetAMotorParam(joint[2],dParamStopERP3,0.3);
		dJointSetAMotorParam(joint[2],dParamStopCFM3,0.05);    */

	joint[0] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[0],body[0],body[1]);
	dJointSetHingeAnchor(joint[0],x,y,z+LEG1[2]/5);
	dJointSetHingeAxis(joint[0],1,0,0);
	dJointSetHingeParam(joint[0],dParamLoStop,-.4); //.8
	dJointSetHingeParam(joint[0],dParamHiStop,.4);  //.4
	dJointSetHingeParam(joint[0],dParamVel,0);
	dJointSetHingeParam(joint[0],dParamFMax,0.5);
	dJointSetHingeParam(joint[0],dParamStopERP,0.4);
	dJointSetHingeParam(joint[0],dParamStopCFM,0.005);

	joint[1] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[1],body[0],body[2]);
	dJointSetHingeAnchor(joint[1],x,y,z+LEG1[2]/5);
	dJointSetHingeAxis(joint[1],1,0,0);
	dJointSetHingeParam(joint[1],dParamLoStop,-.4);
	dJointSetHingeParam(joint[1],dParamHiStop,.4);
	dJointSetHingeParam(joint[1],dParamVel,0);
	dJointSetHingeParam(joint[1],dParamFMax,0.5);
	dJointSetHingeParam(joint[1],dParamStopERP,0.4);
	dJointSetHingeParam(joint[1],dParamStopCFM,0.005);


	// right middle leg
	joint[2] = dJointCreateSlider(world,jGroup);
	dJointAttach(joint[2],body[1],body[3]);
//	dJointSetSliderAnchor(joint[1],TORSO[0]/2+x,y-TORSO[1]/2,z-TORSO[2]/2-LEG1[2]-GAP/2)
	dJointSetSliderAxis(joint[2],0,0,1);
	dJointSetSliderParam(joint[2],dParamLoStop,-0.2);
	dJointSetSliderParam(joint[2],dParamHiStop,0);
	dJointSetSliderParam(joint[2],dParamStopERP,0.3);
	dJointSetSliderParam(joint[2],dParamStopCFM,0.005);
	dJointSetSliderParam(joint[2],dParamBounce,0.8);
	dJointSetSliderParam(joint[2],dParamFudgeFactor,1);
	dJointSetSliderParam(joint[2],dParamVel,0);
	dJointSetSliderParam(joint[2],dParamFMax,0.5);

	joint[3] = dJointCreateSlider(world,jGroup);
	dJointAttach(joint[3],body[2],body[4]);
//	dJointSetSliderAnchor(joint[1],TORSO[0]/2+x,y-TORSO[1]/2,z-TORSO[2]/2-LEG1[2]-GAP/2)
	dJointSetSliderAxis(joint[3],0,0,1);
	dJointSetSliderParam(joint[3],dParamLoStop,-0.2);
	dJointSetSliderParam(joint[3],dParamHiStop,0);
	dJointSetSliderParam(joint[3],dParamStopERP,0.3);
	dJointSetSliderParam(joint[3],dParamStopCFM,0.005);
	dJointSetSliderParam(joint[3],dParamBounce,0.8);
	dJointSetSliderParam(joint[3],dParamFudgeFactor,1);
	dJointSetSliderParam(joint[3],dParamVel,0);
	dJointSetSliderParam(joint[3],dParamFMax,0.5);

}

Hopper::~Hopper()
{
	// destroy joint group
	dJointGroupDestroy(jGroup);
	// destroy Geometries
	for (int i=0;i<numGeom;i++) {
//dGeomGroupRemove(geomGroup,geom[i]);
		dGeomDestroy(geom[i]);
	}
	// destroy bodies
	for (int i=0;i<numBody;i++)
		dBodyDestroy(body[i]);

	// deallocate space
	delete [] body;
	delete [] geom;
	delete [] joint;
}


void Hopper::draw(bool selected) const
{
	dsSetColor (0,1,1);
	dsSetTexture (DS_WOOD); // DS_WOOD
	// head
	dsDrawBox(dBodyGetPosition(body[0]),dBodyGetRotation(body[0]),TORSO);
	if (selected)
		dsSetColor (1,0,0);
	else
		dsSetColor (1,0,1);
	// legs
	dsDrawBox(dBodyGetPosition(body[1]),dBodyGetRotation(body[1]),LEG1);
	dsDrawBox(dBodyGetPosition(body[2]),dBodyGetRotation(body[2]),LEG1);
	dsDrawBox(dBodyGetPosition(body[3]),dBodyGetRotation(body[3]),LEG2);
	dsDrawBox(dBodyGetPosition(body[4]),dBodyGetRotation(body[4]),LEG2);
}

void Hopper::shutoff(void)
{
	dJointSetHingeParam(joint[0],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[0],dParamFMax, (dReal)0);
	dJointSetHingeParam(joint[1],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[1],dParamFMax, (dReal)0);
	dJointSetSliderParam(joint[2],dParamVel,0);
	dJointSetSliderParam(joint[2],dParamFMax,0);
	dJointSetSliderParam(joint[3],dParamVel,0);
	dJointSetSliderParam(joint[3],dParamFMax,0);
}

void Hopper::setMuscles(const float * inputs)
{
//	clock++;
	//  Leg Socket
//	dJointSetAMotorParam(joint[2],dParamVel,(dReal) inputs[16]*1.55);
//	dJointSetAMotorParam(joint[2],dParamFMax, (dReal)fabs(inputs[17]*1.95));
/*	dJointSetAMotorParam(joint[2],dParamVel2,(dReal) inputs[0]);
	dJointSetAMotorParam(joint[2],dParamFMax2,1.05);
	dJointSetAMotorParam(joint[2],dParamVel3,(dReal) inputs[1]);
	dJointSetAMotorParam(joint[2],dParamFMax3,1.05);  */

	dJointSetHingeParam(joint[0],dParamVel,(dReal) inputs[0]*1.5);
	dJointSetHingeParam(joint[0],dParamFMax, (dReal)2.05);
	dJointSetHingeParam(joint[1],dParamVel,(dReal) inputs[0]*1.5);
	dJointSetHingeParam(joint[1],dParamFMax, (dReal)2.05);

	dJointSetSliderParam(joint[2],dParamVel, (dReal) inputs[1]*2.4);
	dJointSetSliderParam(joint[2],dParamFMax, (dReal)6.55);
	dJointSetSliderParam(joint[3],dParamVel, (dReal) inputs[1]*2.4);
	dJointSetSliderParam(joint[3],dParamFMax, (dReal)6.55);
/*	if (hopcycle <= 20)	
		dJointSetSliderParam(joint[1],dParamVel,-10);
	else if (hopcycle > 20 && hopcycle < 40)
		dJointSetSliderParam(joint[1],dParamVel,10);
	else
		hopcycle=0;
	dJointSetSliderParam(joint[1],dParamFMax,6.55);
	hopcycle++;
	dJointSetSliderParam(joint[1],dParamFMax,4.55);
	if (clock < 3) {
		dJointSetSliderParam(joint[1],dParamVel,-4);
	} else if (clock >= 3 && clock < 6) {
		dJointSetSliderParam(joint[1],dParamVel,4);
	}	else clock = 0;   */

}

void Hopper::getInputs(float * inputs) const
{
	// position inputs (position & rotation of head)
	const dReal * vel = dBodyGetLinearVel(body[0]);
  const dReal * vel1 = dBodyGetAngularVel(body[0]);
  const dReal * pos = dBodyGetPosition(body[0]);
	const dReal * rot0 = dBodyGetRotation(body[0]);
  const dReal * rot1 = dBodyGetRotation(body[3]);
  const dReal * pos2 = dBodyGetPosition(body[3]);
  const dReal * pos3 = dBodyGetPosition(body[4]);
//	inputs[0] = (float)pos[1];		// pos y (height) of head
	inputs[0] = (float)pos[2];		// pos z (height) of head
//	inputs[2] = (float)pos2[1];		// pos y foot
//	inputs[3] = (float)pos2[2];    // pos z foot
	inputs[1] = (float)vel[1];    // vel y
  inputs[2] = (float)vel[2];    // vel z
  inputs[3] = (float)rot0[2];
  inputs[4] = (float)rot1[2];   // z rotation of foot
	inputs[5] = (float)vel1[2];   // z angular velocity of body
//	inputs[7] = (float)pos[1];
//	inputs[8] = (float)pos[2];
	// get right leg socket angles (3)
/*	inputs[5] = (float)dJointGetAMotorAngle(joint[2],0); */
//	inputs[3] = (float)dJointGetAMotorAngle(joint[2],1);
//	inputs[4] = (float)dJointGetAMotorAngle(joint[2],2);	// get left leg socket angles (3)
	inputs[6] = (float)dJointGetSliderPosition(joint[1]);

//	inputs[6] = (float)dJointGetHingeAngle(joint[0]);
	inputs[7] = (float)dJointGetHingeAngle(joint[1]);
	if (r_foot_sens)
		inputs[8] = 1.0;
	else	
		inputs[8] = -1.0;
/*	if (l_foot_sens)
		inputs[9] = 1.0;
	else	
		inputs[9] = -1.0; */
}

void Hopper::bodyContact(dGeomID o1, dGeomID o2, dGeomID ground)
{
	if (((o1==geom[0]) && (o2==ground)) || ((o2==geom[0])&&(o1==ground)))
		fallen=true;
  if (!fallen)
    fitcount += 0.01;
}

void Hopper::footContact(dGeomID o1, dGeomID o2, dGeomID ground)
{
	if (((o1==geom[3]) && (o2==ground)) || ((o2==geom[3])&&(o1==ground)))
		r_foot_sens=true;
	if (((o1==geom[4]) && (o2==ground)) || ((o2==geom[4])&&(o1==ground)))
    l_foot_sens=true;
}

float Hopper::getFitness(void) const
{

	const dReal * pos = getHeadPos();

	float fitness = -pos[1]*9 - pow(fabs(getStartX()-pos[0]),4);
  if (fitness > 0)
		return fitness;
	else
		return 0;		// y coordinate.
	// fitness function
/*	if (fabs(pos[1]) > 30 || (fabs(pos[2]) > 10) || (fabs(pos[0]-getStartX()) > 10))
		fitness = 0;
	else
		fitness = -pos[1]*9+fitcount;

  if (!(fitness > 0 && fitness < 200))
    fitness = 0; */

	return fitness;
}

const dReal *Hopper::getHeadPos(void) const
{
	return dBodyGetPosition(body[0]);
}

void Hopper::randomForce(void) const
{
    dReal f[3];
    f[0] = 0.55; //* (-0.5+1.0/RAND_MAX * rand());
    f[1] = 2.75 * (-0.5+1.0/RAND_MAX * rand());
    f[2] = 0.3 * (-0.5+1.0/RAND_MAX * rand());
    dBodyAddForce(body[0], f[0], f[1], f[2]);
}
void Hopper::randomTorque(void) const
{
    dReal f[3];
    f[0] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
    f[1] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
    f[2] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
	  dBodyAddTorque(body[0], f[0], f[1], f[2]);
}

bool Hopper::hasFallen(void) const
{
	return fallen;
}

void Hopper::printStats(void) const
{
	const dReal *pos = dBodyGetPosition(body[0]);
	const dReal *rot = dBodyGetRotation(body[0]);
	const dReal *vel = dBodyGetLinearVel(body[0]);

	cout << "\nPosition:";
	cout << "X: " << pos[0] << " ";
	cout << "Y: " << pos[1] << " ";
	cout << "Z: " << pos[2] << endl;
	cout << "Rotation:" << endl;
	cout << "XR: " << rot[0] << " ";
	cout << "YR: " << rot[1] << " ";
	cout << "ZR: " << rot[2] << endl;
	cout << "Vel:" << endl;
	cout << "Vx: " << vel[0] << " ";
	cout << "Vy: " << vel[1] << " ";
	cout << "Vz: " << vel[2] << endl;
	cout << "l_foot: " << l_foot_sens << "  r_foot: " << r_foot_sens << endl;
}

void Hopper::clearSensor(void)
{
  r_foot_sens=false;
  l_foot_sens=false;
}

void Hopper::Reset(double x, double y, double z)
{
  dMatrix3 R;
  dQuaternion q;
  dRSetIdentity(R);
  dQSetIdentity(q);

	dBodySetPosition (body[0],0+x,0+y,z);	// torso
	dBodySetPosition (body[1],x-TORSO[0]/4,y,z-LEG1[2]/5-GAP);	//right upper leg
	dBodySetPosition (body[2],x+TORSO[0]/4,y,z-LEG1[2]/5-GAP);	//left upper leg
	dBodySetPosition (body[3],x-TORSO[0]/4,y,z-LEG1[2]-LEG2[2]/2-GAP);	//right foot
	dBodySetPosition (body[4],x+TORSO[0]/4,y,z-LEG1[2]-LEG2[2]/2-GAP);	//left foot

	for (int i=0;i<numBody;i++)
  {
    dBodySetRotation(body[i],R);
    dBodySetQuaternion(body[i],q);
    dBodySetLinearVel(body[i],0,0,0);
    dBodySetAngularVel(body[i],0,0,0);
  }

  for (int i=0;i<=1;i++)
  {
	  dJointSetHingeParam(joint[0],dParamVel,0);
	  dJointSetHingeParam(joint[0],dParamFMax,0.5);
  }

  for (int i=2;i<=3;i++)
  {
	  dJointSetSliderParam(joint[2],dParamVel,0);
	  dJointSetSliderParam(joint[2],dParamFMax,0.5);
  }
  fallen = false;
  fitcount=0.0;
}