/***************************************************************************
                          main.cpp  -  description
                             -------------------
    begin                : Mon Feb 9 2004
    copyright            : (C) 2004 by Gene Ruebsamen
    email                : gene@localhost.localdomain
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <ode/ode.h>
#include <drawstuff/drawstuff.h>
#include <iostream>
#include <cstdlib>
#include <ctime>
#include <iomanip>
#include <fstream>
#include "constants.h"
#include "world.h"
#include "genetic.h"
#include "organisms/organism.h"
#include "organisms/newbox.h"
#include "organisms/paddle.h"
#include "brains/neuralnet.h"
#include "brains/net2.h"
//#include "ann.h"

using std::cout;
using std::endl;
using std::setw;
using std::ios;
using std::cerr;
using std::ofstream;

// select correct drawing functions
#ifdef dDOUBLE
#define dsDrawBox dsDrawBoxD
#define dsDrawSphere dsDrawSphereD
#define dsDrawCylinder dsDrawCylinderD
#define dsDrawCappedCylinder dsDrawCappedCylinderD
#endif

dVector3 gravity={0,0,-9.8};
World earth(gravity);
Genetic genAlg;
dGeomID ground;

vector<Organism<float> *> population;

//----------
// start simulation - set viewpoint
//----------
static void start()
{
	static float xyz[3] = {0.8317f,-0.9817f,1.8000f};
  static float hpr[3] = {121.0000f,-27.5000f,0.0000f};
  dsSetViewpoint (xyz,hpr);
  printf ("Press:\t'a' to increase speed.\n"
	  "\t'z' to decrease speed.\n"
	  "\t',' to steer left.\n"
	  "\t'.' to steer right.\n"
	  "\t' ' to reset speed and steering.\n"); 
}

//----------
// Respond to Keypresses during simulation
//----------
static void command (int cmd)
{
/*
  switch (cmd) {
  case 'n': case 'N' :
		if (display < NUMHUMANS-1)
			display++;
		else
			display=0;
	break;
  case 'p': case 'P' :
		if (display > 0)
			display--;
		else
			display=NUMHUMANS-1;
	break;
	case 'i': case 'I' :
  maxtime +=50;
  cout << "MaxTime: " << maxtime << endl;
	break;
	case 'd': case 'D' :
  maxtime -=50;
  cout << "MaxTime: " << maxtime << endl;
	break;

  case 'm': case 'M' :
    human[display]->printStats();
  break;
  }
  */   
}

// run Simulation
static void simLoop(int pause)
{        
  static int timer = 0;
  static unsigned short int count = population[0]->getResponseTime();
  if (timer < TIMEOUT)
  {
		if (count == 0)
		{
			//population[0]->printStats();
			for (int i=0;i<population.size();++i)
			{
				population[i]->groundContact(ground);
				vector<float> data = population[i]->getSensoryData();
				//vector<float> data2 = data;
				//for (int j=0;j<data.size();++j)
					//data2.push_back(data[j]);
					
				//population[i]->activateMotors(population[i]->getBrainOutput(data2));
				population[i]->activateMotors(population[i]->getBrainOutput(data));
				//population[i]->getBrainOutput(data2);
			}
			count = population[0]->getResponseTime();
		} else {
			count--;
		}
		earth.fStep(SIMSTEPMS);
		// draw
		dsSetColor(0.7,0.3,0.7);
		for (int i=0;i<population.size();++i)
		{
			population[i]->draw(false);
		}
		timer++;
	} 
	else  // do ga
	{
		// calculate fitness of all individuals
		for (int i=0;i<population.size();++i)
		{
			population[i]->setFitness(population[i]->calcFitness());
		}
		// create new population

		vector<Organism<float> *> newPopulation = genAlg.epoch(population);
		
		genAlg.printStats();
		// modify weights
		for (int i=0;i<population.size();++i)
		{
			delete population[i];
		}
		population = newPopulation;//new NewBox(earth.getWorldID(),earth.getSpaceID(),start);

		// reset timer
		timer = 0;
	}
}	// end simLoop Function

// main program
int main(int argc, char *argv[]) {

	// setup pointers to drawstuff callback functions
	dsFunctions fn;
	fn.version = DS_VERSION;
	fn.start = &start;
	fn.step = &simLoop;
	fn.command = &command;
	fn.stop = 0;
	fn.path_to_textures = "/home/gene/thesis/textures";
	srand(time(0));			// seed random number generator

	// create world (setup environment)
	dWorldSetGravity (earth.getWorldID(),0,0,-GRAVITY);	// set gravity (x,y,z)

	// create the ground plane
  ground = dCreatePlane(earth.getSpaceID(),0,0,1,0);
 
	// create creatures forward
	for (int i=0;i<POPSIZE;++i)
	{
		dVector3 start = {(i*SPACING)-POPSIZE/2,0,STARTZ};
		population.push_back(new Paddle<float>(earth.getWorldID(),earth.getSpaceID(),start));
	}
  genAlg.setNumElite(5);
	// Start Simulation
	dsSimulationLoop (argc,argv,640,480,&fn);	// 640 x 480 OpenGL display
	
	for (int i=0;i<population.size();++i)
		delete population[i];
		
	dCloseODE();
}  /*
	double **weights;
	weights = new double*[1000];
	for (int i=0;i<1000;++i)
		weights[i] = new double[1000];

	int **firings;
	firings = new int*[1000];
	for (int i=0;i<1000;++i)
		firings[i] = new int[1000];

	// fill weights
	for (int i=0;i<1000;++i)
	{
		for (int j=0;j<800;++j)
		{
			weights[i][j] = RandFloat()*0.5;
		}
		for (int j=800;j<1000;++j)
		{
			weights[i][j] = -RandFloat();
		}
	}
	cout << "here" << endl;
	vector<SpikingNeuron<double> *> neurons;
	// exhibitory
	for (int i=0;i<800;++i)
		neurons.push_back(new SpikingNeuron<double>(0));
	// inhibitory
	for (int i=0;i<200;++i)
		neurons.push_back(new SpikingNeuron<double>(4));
	// experimental
	for (int time=0;time<1000;++time)
	{
		vector<double> inputs;
		for (int i=0;i<800;++i)
			inputs.push_back(5.0*(RandClamped()*2.0));
		for (int i=0;i<200;++i)
			inputs.push_back(2.0*(RandClamped()*2.0));
		// find firings
		for (int i=0;i<neurons.size();++i)
		{
			if (neurons[i]->findFired())
			{
				firings[i][time] = 1;
				for (int j=0;j<1000;++j)
					inputs[j]+=weights[i][j];
			} else {
				firings[i][time] = 0;
			}
		}
		// add I to neurons
		for (int i=0;i<inputs.size();++i)
		{
			neurons[i]->receiveSpike(inputs[i]);
			neurons[i]->step(1.0);
		}
	}

	for (int time=0;time<1000;++time)
	{
		for (int i=0;i<1000;++i)
		{
			cout << firings[i][time] << " ";
		}
		cout << endl;
	}
}
/*
	// EXPERIMENTAL STUFF BELOW:
	vector<float> input;
	SpikeNet<float> *temp = new SpikeNet<float>(1000,0,1000);
	for (int i=0;i<1000;++i)
	{
		for (int i=0;i<800;++i)
			input.push_back(5.0*(RandClamped()*2.0));
		for (int i=0;i<200;++i)
			input.push_back(2.0*(RandClamped()*2.0));
		temp->activate(input);
		temp->step(1);
//		cout << "t: " << i << "  Firings: " << temp->firecount << endl;
		temp->print(0,1000);
		input.clear();
	}
}   

     /*
	// 2
	SpikingNeuron<double> *a= new SpikingNeuron<double>(3);
	SpikingNeuron<double> *a1= new SpikingNeuron<double>(3);
	SpikingNeuron<double> *b= new SpikingNeuron<double>(0);
	SpikingNeuron<double> *c= new SpikingNeuron<double>(0);
  a->setInput(0);
  a1->setInput(0);
  b->setInput(0);
  c->setInput(0);

  // connect a->b a->c
  Connection<double> con1(9.5,b);
  Connection<double> con2(9.5,c);
  
  a->addConnection(con1);
  a->addConnection(con2);
  a1->addConnection(con1);
  a1->addConnection(con2);
  
	for (int i=0;i<150;++i)
	{
		a->step(0.1);
		a1->step(0.1);
		b->step(0.1);
		c->step(0.1);
		cout << a->printPotential() << " " << b->printPotential() << " " << c->printPotential() << endl;
	}
	// set input to 10 mV
	a->setInput(10.0);
	a1->setInput(10.0);
	for (int i=0;i<3000;++i)
	{
		a->step(0.1);
		a1->step(0.1);
		double spike = 4.5;
		//if (i%10 == 0) {
		//	for (int k=0;k<5;++k)
		//		b->receiveSpike(spike);
		//}
		b->step(0.1);
		c->step(0.1);
		cout << a->printPotential() << " " << b->printPotential() << " " << c->printPotential() << endl;
	}
	// END EXPERIMENTAL
  return 0;
}      */
 
               
               