/***************************************************************************
                          net.cpp  -  description
                             -------------------
    begin                : Sat Dec 1 2001
    copyright            : (C) 2001 by Gene Ruebsamen
    email                : gene@erachampion.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <cmath>
#include <cstdlib>
#include "net.h"

Net::Net(int in, int hid, int out, int recur) : input_size(in), hidden_size(hid), output_size(out), recurrent_size(recur)
{
	time = 0;
	whi = new float[hidden_size*input_size];		// input hidden weights
	whr = new float*[hidden_size*hidden_size];	// recurrent layer
	for (int i=0;i<hidden_size*hidden_size;i++)
		whr[i] = new float[recurrent_size];
	whb = new float[hidden_size];
	woi = new float[output_size*hidden_size];
	wob = new float[output_size];
	// outputs
	ho = new float[hidden_size];			// hidden layer outputs
	oo = new float[output_size];			// output layer outputs
	hiddencopy = new float*[hidden_size];		// for feedback
	for (int i=0;i<hidden_size;i++)
		hiddencopy[i] = new float[recurrent_size];
}

Net::Net(const Net &copy) : input_size(copy.input_size), hidden_size(copy.hidden_size), output_size(copy.output_size), recurrent_size(copy.recurrent_size)
{
	time = 0;
	whi = new float[hidden_size*input_size];
	whr = new float*[hidden_size*hidden_size];	// recurrant layer
	for (int i=0;i<hidden_size*hidden_size;i++)
		whr[i] = new float[recurrent_size];
	whb = new float[hidden_size];
	woi = new float[output_size*hidden_size];
	wob = new float[output_size];
	// outputs
	ho = new float[hidden_size];			// hidden layer outputs
	oo = new float[output_size];			// output layer outputs
	hiddencopy = new float* [hidden_size];		// for feedback
	for (int i=0;i<hidden_size;i++)
		hiddencopy[i] = new float[recurrent_size];

	for (int i=0;i<hidden_size;i++)
	{
		whb[i] = copy.whb[i];
		ho[i] = copy.ho[i];
	}

	// recurrent weights (hidden*hidden)
	for (int i=0;i<recurrent_size;i++)
		for (int j=0;j<hidden_size*hidden_size;j++) {
			whr[j][i] = copy.whr[j][i];
		}

	// hidden copy layer (size = hidden)
	for (int i=0;i<recurrent_size;i++)
		for (int j=0;j<hidden_size;j++) {
			hiddencopy[j][i] = copy.hiddencopy[j][i];
		}

	for (int i=0;i<output_size;i++)
	{
		wob[i] = copy.wob[i];
		oo[i] = copy.oo[i];
	}
	for (int i=0;i<(hidden_size*input_size);i++)
		whi[i] = copy.whi[i];
	for (int i=0;i<(output_size*hidden_size);i++)
		woi[i] = copy.woi[i];
}

const Net &Net::operator=(const Net &right)
{
	time = 0;
	if (&right != this)	{			// check for self-assignment
		if ((input_size != right.input_size) || (hidden_size != right.hidden_size) || (output_size != right.output_size)
				|| (recurrent_size != right.recurrent_size)) {
			delete [] ho; delete [] oo;
  		delete [] whb; delete [] whi;  //delete [] whr2;
			delete [] wob; delete [] woi; //delete [] wib;
			for (int i=0;i<hidden_size;i++)
				delete [] hiddencopy[i];
			for (int i=0;i<hidden_size*hidden_size;i++)
				delete [] whr[i];

			input_size = right.input_size;
			hidden_size = right.hidden_size;
			output_size = right.output_size;
			whr = new float* [hidden_size*hidden_size];	// recurrant layer
			for (int i=0;i<hidden_size*hidden_size;i++)
				whr[i] = new float[recurrent_size];
			whi = new float[hidden_size*input_size];
			whb = new float[hidden_size];
			woi = new float[output_size*hidden_size];
			wob = new float[output_size];
			// outputs
			ho = new float[hidden_size];			// hidden layer outputs
			oo = new float[output_size];			// output layer outputs
			hiddencopy = new float* [hidden_size];		// for feedback
			for (int i=0;i<hidden_size;i++)
				hiddencopy[i] = new float[recurrent_size];
		}
	}
	// copy items over for net
	for (int i=0;i<hidden_size;i++)
	{
		whb[i] = right.whb[i];
		ho[i] = right.ho[i];
	}

	// copy recurrent weights
	for (int i=0;i<recurrent_size;i++)
		for (int j=0;j<hidden_size*hidden_size;j++) {
			whr[j][i] = right.whr[j][i];
		}

	// copy hidden layers
	for (int i=0;i<recurrent_size;i++)
		for (int j=0;j<hidden_size;j++) {
			hiddencopy[j][i] = right.hiddencopy[j][i];
		}

	for (int i=0;i<output_size;i++)
	{
		wob[i] = right.wob[i];
		oo[i] = right.oo[i];
	}
	for (int i=0;i<(hidden_size*input_size);i++)
		whi[i] = right.whi[i];
	for (int i=0;i<(output_size*hidden_size);i++)
		woi[i] = right.woi[i];
	return *this;
}

void Net::initWeights(float *weights)
{
	int i;
	// setup hidden weight array
	for (i=0;i<(input_size*hidden_size);i++)
		whi[i] = *weights++;
	// setup hidden recurrent weights
	for (int i=0;i<recurrent_size;i++)
		for (int j=0;j<hidden_size*hidden_size;j++)
			whr[j][i] = *weights++;
	// setup hidden bias
	for (i=0;i<hidden_size;i++)
		whb[i] = *weights++;
	// setup output weight array
	for (i=0;i<(hidden_size*output_size);i++)
		woi[i] = *weights++;
	// setup output bias
	for (i=0;i<(output_size);i++)
		wob[i] = *weights++;
}

Net::~Net()
{
	delete [] ho;  delete [] oo;
	delete [] whb; delete [] whi; //delete [] whr2;
  delete [] wob; delete [] woi;
	for (int i=0;i<hidden_size;i++)
		delete [] hiddencopy[i];
	for (int i=0;i<hidden_size*hidden_size;i++)
		delete [] whr[i];
}

const float * const Net::getOutput(const float * const input)
{
	//static int time=0;
	// calculate the output from the hidden layer
	for (int i=0;i<hidden_size;i++)
	{
		float sum=0.0;
		for (int j=0;j<input_size;j++) {
			sum += (input[j]*whi[j+(i*input_size)]);
		}

		if (time > 0) {
			for (int k=0;k<time;k++)
				for (int m=0;m<hidden_size;m++) {
					sum += (hiddencopy[m][k]*whr[i*hidden_size+m][k]);
				}
		}

		sum += whb[i];												// bias term
		ho[i] = (2.0/(1.0+exp(-2*sum)))-1;         	// use bipolar sigmoid activation function
		hiddencopy[i][0] = ho[i];                		// copy hidden layer
	}

	if (time >= 1) {
		for (int i=recurrent_size-2;i>=0;i--)					// copy hidden layers
			for (int j=0;j<hidden_size;j++)
				hiddencopy[j][i+1] = hiddencopy[j][i];
	}
	// calculate the output from the output layer
	for (int i=0;i<output_size;i++)
	{
		float sum=0.0;
		for (int j=0;j<hidden_size;j++)
		{
			sum += (ho[j]*woi[j+(i*hidden_size)]);
		}
		sum += wob[i];														// bias term
    oo[i] = (2.0/(1.0+exp(-2*sum)))-1;
	}
	if (time < recurrent_size)
		time++;
	return oo;																	// return output
}

const int Net::getSize(void) const
{
	return ((input_size*hidden_size) + (hidden_size*output_size)
					+ (hidden_size+output_size) + (hidden_size*hidden_size*recurrent_size));
}


