/***************************************************************************
                          newbox.h  -  description
                             -------------------
    begin                : Wed Feb 11 2004
    copyright            : (C) 2004 by Gene Ruebsamen
    email                : gene@localhost.localdomain
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef NEWBOX_H
#define NEWBOX_H
#include <ode/ode.h>
#include <drawstuff/drawstuff.h>
#include "organism.h"
#include "../brains/neuralnet.h"
#include "../brains/brain.h"

template <class T>
class NewBox : public Organism<T> {
public:
	NewBox(dWorldID,dSpaceID,dVector3,bool=true);		// constructor (world,space,x,y,z,in,hid,out,recur)
	NewBox(const NewBox &);
	~NewBox();	// destructor

	// cognitive functions
	vector<T> getSensoryData() const;
	inline vector<T> getBrainOutput(vector<T> &input);
	void activateMotors(const vector<T> &);

	// GA Functions
	Organism<T> *operator+(const Organism<T> &right);		// reproduction
	const NewBox &operator=(const NewBox &right);		// assignment operator

	// Utility functions
	void groundContact(dGeomID);
	const dReal *getBodyRot() const;
	const double calcFitness() const;
	inline const dReal *getCurrentPosition() const;
	void setStartPosition(dVector3);
	void footContact(dGeomID,dGeomID,dGeomID);   		// are the feet touching the ground
	void enable(bool);
	void reset();
	void printStats() const;
	void draw(bool=false) const;
  
private:
	void constructBody(const dWorldID,const dSpaceID,const dVector3);
	int numBody;
	int numGeom;
	int numJoint;
	bool l_foot_sens;
	bool r_foot_sens;
	dBodyID *body;
	dGeomID *geom;
	dJointID *joint;
	dJointGroupID jGroup;
	dGeomID geomGroup;
	//  declarations
	dReal TORSO[3]; 	// length, width height
	dReal LEG1[3];
	dReal LEG2[3];
	dReal LEG3[3];
	dReal GAP;
};

#endif
