/***************************************************************************
                          organism.cpp  -  description
                             -------------------
    begin                : Mon Feb 9 2004
    copyright            : (C) 2004 by Gene Ruebsamen
    email                : gene@localhost.localdomain
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "organism.h"

template class Organism<float>;
template class Organism<double>;

template <class T>
Organism<T>::Organism(dWorldID worldid, dSpaceID spaceid, dVector3 position, bool enabled) :
m_Fitness(0.0),
m_Selected(false),
m_ResponseTime(1)
{
	m_WorldID = worldid;
	m_SpaceID = spaceid;
	m_Enabled = enabled;
	for (unsigned i=0;i<sizeof(position)/sizeof(dReal);++i)
		m_StartPosition[i] = position[i];
}

template <class T>
Organism<T>::Organism(const Organism &copy)
{
	m_WorldID = copy.m_WorldID;
	m_Enabled = copy.m_Enabled;
	m_SpaceID = copy.m_SpaceID;
	m_Fitness = copy.m_Fitness;
	m_ResponseTime = copy.m_ResponseTime;
	m_Selected = copy.m_Selected;
	for (int i=0;i<sizeof(m_StartPosition)/sizeof(dReal);++i)
		m_StartPosition[i] = copy.m_StartPosition[i];
}

template <class T>
Organism<T>::~Organism()
{
}

template <class T>
const dReal *Organism<T>::getStartPosition() const
{
	return m_StartPosition;
}

template <class T>
void Organism<T>::setStartPosition(dVector3 pos)
{
	for (int i=0;i<sizeof(m_StartPosition)/sizeof(dReal);++i)
	{
		m_StartPosition[i] = pos[i];
	}
}

template <class T>
void Organism<T>::enable(bool enable)
{
	m_Enabled = enable;
}

template <class T>
bool Organism<T>::isEnabled() const
{
	return m_Enabled;
}

// GA Operators
template <class T>
const Organism<T> &Organism<T>::operator=(const Organism<T> &right)
{
	if (&right != this)			// avoid self-assignment
	{
		m_WorldID = right.m_WorldID;
		m_Enabled = right.m_Enabled;
		m_SpaceID = right.m_SpaceID;
		m_Fitness = right.m_Fitness;
		m_ResponseTime = right.m_ResponseTime;
		delete m_Brain;
		*m_Brain = *(right.m_Brain);
		for (unsigned i=0;i<sizeof(m_StartPosition)/sizeof(dReal);++i)
			m_StartPosition[i] = right.m_StartPosition[i];
	}

	return *this;
}