/***************************************************************************
                          organism.h  -  description
                             -------------------
    begin                : Mon Feb 9 2004
    copyright            : (C) 2004 by Gene Ruebsamen
    email                : gene@localhost.localdomain
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ORGANISM_H
#define ORGANISM_H
#include <ode/ode.h>
#include "../brains/brain.h"
#include "../constants.h"
#include "../utils.h"

template <class T>
class Organism {
public:
	Organism(dWorldID, dSpaceID, dVector3 position, bool enabled=true);
	Organism(const Organism &);
	virtual ~Organism();	// destructor
	virtual void draw(bool=true) const=0;

	// Misc functions
	virtual void groundContact(dGeomID ground)
		{ }
	virtual void enable(bool);
	void select(bool selected=false)
		{ m_Selected = selected; }
	virtual void reset()=0;
	inline bool isEnabled() const;
	bool isSelected() const
		{ return m_Selected; }
	unsigned short int getResponseTime() const
		{ return m_ResponseTime; }
	void setResponseTime(unsigned short int time)
		{ m_ResponseTime = time; }
	const dReal *getStartPosition() const;
	virtual void setStartPosition(dVector3);
	virtual const dReal *getCurrentPosition() const=0;
	virtual void printStats() const=0;
	dWorldID getWorldID() const
		{ return m_WorldID; }
	dSpaceID getSpaceID() const
		{ return m_SpaceID; }

	// Cognitive functions
	virtual vector<T> getSensoryData() const=0;
	virtual vector<T> getBrainOutput(vector<T> &input)=0;
	virtual void activateMotors(const vector<T> &)=0;

	// GA functions
	double getFitness() const
		{ return m_Fitness; }
	void setFitness(double fit)
		{ m_Fitness = fit; }
	virtual const double calcFitness() const=0;
	void setChromosome(vector<T> &in)
		{ m_Brain->putWeights(in); }
	vector<T> getChromosome() const
		{	return m_Brain->getWeights(); }
	virtual Organism *operator+(const Organism &right)=0;		// reproduction
	const Organism &operator=(const Organism &right);		// assignment operator

protected:
	Brain<T> *m_Brain;
	dWorldID m_WorldID;
	dSpaceID m_SpaceID;
	dVector3 m_StartPosition;
	
private:
	double m_Fitness;
	bool m_Enabled;
	bool m_Selected;
	unsigned short int m_ResponseTime;
};

#endif

