/***************************************************************************
                          paddle.cpp  -  description
                             -------------------
    begin                : Wed Feb 18 2004
    copyright            : (C) 2004 by Gene Ruebsamen
    email                : gene@localhost.localdomain
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "paddle.h"      // import constants for bodies
#include <iostream>

using std::cout;
using std::endl;

template class Organism<float>;
template class Organism<double>;
template class Paddle<float>;
template class Paddle<double>;

template <class T>
Paddle<T>::Paddle(dWorldID world, dSpaceID space, dVector3 position, bool enabled) :
Organism<T>(world,space,position,enabled),
numBody(3),
numGeom(3),
numJoint(4)
{
	// Brain Setup
	//m_Brain = new SpikeNet<T>(22,6,22);
	m_Brain = new NeuralNet<T>(11,8,6,8,1);
	//m_Brain = new Net2<T>(11,8,6,8);
	setResponseTime(5);
	constructBody(world,space,position);
}


template <class T>
Paddle<T>::~Paddle()
{
	// destroy joint group
	dJointGroupDestroy(jGroup);
	for (int i=0;i<numGeom;i++) {
		dGeomDestroy(geom[i]);
	}
	// destroy bodies
	for (int i=0;i<numBody;i++)
		dBodyDestroy(body[i]);

	// deallocate space
	delete m_Brain;
	delete [] body;
	delete [] geom;
	delete [] joint;
}


template <class T>
Paddle<T>::Paddle(const Paddle<T> &right) :
Organism<T>(right)
{
	numBody = right.numBody;
	numGeom = right.numGeom;
	numJoint = right.numJoint;

	//SpikeNet<T> *temp = dynamic_cast<SpikeNet<T> *>(right.m_Brain);
	//m_Brain = new SpikeNet<T>(*temp);
	
	NeuralNet<T> *temp = dynamic_cast<NeuralNet<T> *>(right.m_Brain);
	//Net2<T> *temp = dynamic_cast<Net2<T> *>(right.m_Brain);
	m_Brain = new NeuralNet<T>(*temp);
	//m_Brain = new Net2<T>(*temp);
	constructBody(right.m_WorldID, right.m_SpaceID, right.m_StartPosition);
}

template <class T>
void Paddle<T>::reset()
{
	// reset the foot sensors
  r_foot_sens=false;
  l_foot_sens=false;

  //SpikeNet<T> *brain_ptr = dynamic_cast<SpikeNet<T> *>(m_Brain);
  
  NeuralNet<T> *brain_ptr = dynamic_cast<NeuralNet<T> *>(m_Brain);
  //Net2<T> *brain_ptr = dynamic_cast<Net2<T> *>(m_Brain);
  int output_size = brain_ptr->getOutputSize();
  vector<T> motor_outputs;
  for (int i=0;i<output_size;++i)
  {
		motor_outputs.push_back(0);
  }
  // reset the memory
  brain_ptr->clearMemory();
  // reset the motors to zero
  activateMotors(motor_outputs);
  // reset the forces
  dQuaternion q = {0,0,0,0};
	for (int i=0;i<numBody;i++)
	{
		dBodySetLinearVel(body[i],0,0,0);
		dBodySetAngularVel(body[i],0,0,0);
		dBodySetQuaternion(body[i],q);
	}
 
}

template <class T>
void Paddle<T>::draw(bool selected) const
{
	dsSetColor (0,1,1);
	dsSetTexture (DS_WOOD); // DS_WOOD
	// head
	dsDrawBox(dBodyGetPosition(body[0]),dBodyGetRotation(body[0]),TORSO);
	if (selected || isSelected())
		dsSetColor (1,0,0);
	else
		dsSetColor (1,0,1);
	// legs
	for (int i=1;i<=2;i++)
		dsDrawBox(dBodyGetPosition(body[i]),dBodyGetRotation(body[i]),LEG);
}

template <class T>
void Paddle<T>::activateMotors(const vector<T> &inputs)
{
	// left Arm Socket
	dJointSetAMotorParam(joint[2],dParamVel,(dReal) inputs[0]*1.5);
	dJointSetAMotorParam(joint[2],dParamFMax,(dReal)1);
	dJointSetAMotorParam(joint[2],dParamVel2,(dReal) inputs[1]);
	dJointSetAMotorParam(joint[2],dParamFMax2,(dReal)1);
	dJointSetAMotorParam(joint[2],dParamVel3,(dReal) inputs[2]*1.5);
	dJointSetAMotorParam(joint[2],dParamFMax3,(dReal)1);
	// right Arm socket
	dJointSetAMotorParam(joint[3],dParamVel,(dReal) inputs[3]*1.5);
	dJointSetAMotorParam(joint[3],dParamFMax,(dReal)1);
	dJointSetAMotorParam(joint[3],dParamVel2,(dReal) inputs[4]);
	dJointSetAMotorParam(joint[3],dParamFMax2,(dReal)1);
	dJointSetAMotorParam(joint[3],dParamVel3,(dReal) inputs[5]*1.5);
	dJointSetAMotorParam(joint[3],dParamFMax3,(dReal)1);
}

template <class T>
vector<T> Paddle<T>::getSensoryData() const
{
	vector<T> sensor;
	// position inputs (position & rotation of head)
	const dReal * angle = dBodyGetRotation(body[0]);
	const dReal * vel = dBodyGetLinearVel(body[0]);
  const dReal * pos = dBodyGetPosition(body[0]);

	sensor.push_back(dJointGetAMotorAngle(joint[2],0));		// 0
	sensor.push_back(dJointGetAMotorAngle(joint[2],1));		// 1
	sensor.push_back(dJointGetAMotorAngle(joint[2],2));		// 2
	sensor.push_back(dJointGetAMotorAngle(joint[3],0));		// 3
	sensor.push_back(dJointGetAMotorAngle(joint[3],1));		// 4
	sensor.push_back(dJointGetAMotorAngle(joint[3],2));		// 5

/*	const dReal * bod = dBodyGetPosition(body[0]);
	const dReal * box = dBodyGetPosition(body[3]);

	// direction sensor
	float x = bod[0]-box[0];
	float y = bod[1]-box[1];
	float hyp = sqrt(pow(x,2)+pow(y,2));  */

	//inputs[6] = (float)asin(x/hyp);	// radians
	//inputs[7] = (float)hyp;

	sensor.push_back(angle[1]); // 6
	sensor.push_back(vel[0]);		// 7
//	inputs[0]=(float)fabs(vel[1]) + (float)fabs(vel[0]);
	// foot touch sensor
	if (r_foot_sens)
		sensor.push_back(1.0);	// 8
	else
		sensor.push_back(-1.0);
	if (l_foot_sens)
		sensor.push_back(1.0);	// 9
	else
		sensor.push_back(-1.0);
	if (body_sens)
		sensor.push_back(1.0);	// 10
	else
		sensor.push_back(-1.0);

	return sensor;
}

template <class T>
void Paddle<T>::groundContact(dGeomID ground)
{
	const int N = 5;	// approximate 5 contacts
	dContact contact[N];
	r_foot_sens = false;
	l_foot_sens = false;
	body_sens = false;

	int n = dCollide (geom[1],ground,N,&contact[0].geom,sizeof(dContact));
  // contact!
	if (n > 0) {
		l_foot_sens=true;
	}

	n = dCollide (geom[2],ground,N,&contact[0].geom,sizeof(dContact));
	if (n > 0) {
		r_foot_sens=true;
  }

	n = dCollide (geom[0],ground,N,&contact[0].geom,sizeof(dContact));
	if (n > 0) {
		body_sens=true;
  }

}

template <class T>
const dReal *Paddle<T>::getCurrentPosition() const
{
	return dBodyGetPosition(body[0]);
}

template <class T>
const dReal *Paddle<T>::getBodyRot() const
{
	return dBodyGetRotation(body[0]);
}

template <class T>
const double Paddle<T>::calcFitness() const
{
	const dReal *pos = getCurrentPosition();
  double result = -pos[1]*9 - pow(fabs(m_StartPosition[0]-pos[0]),4);

	if (result > 0)
		return result;
	else
		return 0;		// y coordinate.
}

template <class T>
void Paddle<T>::printStats(void) const
{
	const dReal *pos = dBodyGetPosition(body[0]);
	const dReal *rot = dBodyGetRotation(body[0]);

	cout << "\nPosition:";
	cout << "X: " << pos[0] << " ";
	cout << "Y: " << pos[1] << " ";
	cout << "Z: " << pos[2] << endl;
	cout << "Rotation:" << endl;
	cout << "XR: " << rot[0] << " ";
	cout << "YR: " << rot[1] << " ";
	cout << "ZR: " << rot[2] << endl;
	cout << "Feet:" << endl;
	cout << "Right: " << r_foot_sens << " ";
	cout << "Left: " << l_foot_sens << " ";
	cout << "Body: " << body_sens << endl;
}

template <class T>
vector<T> Paddle<T>::getBrainOutput(vector<T> &input)
{
	return(m_Brain->activate(input));
}

template <class T>
Organism<T> *Paddle<T>::operator+(const Organism<T> &right)
{
  vector<T> leftChromo = m_Brain->getWeights();
  vector<T> rightChromo = dynamic_cast<const Paddle<T> *>(&right)->m_Brain->getWeights();

	int length = m_Brain->getGeneSize();
  vector<T> chromosome;
  
  // Chance of Crossover
	if (RandFloat() <= CROSSOVER)
	{
	  // Determine Crossover Point
	 	int crosspoint = (RandInt(0,length-1));
	  for (int i=0;i<crosspoint;++i)
			chromosome.push_back(leftChromo[i]);
	  for (int i=crosspoint;i<length;++i)
			chromosome.push_back(rightChromo[i]);
	}
	else // just copy
	{
		if (RandFloat() < 0.50)
		{
			for (int i=0;i<length;++i)
				chromosome.push_back(leftChromo[i]);
		}
		else
		{
			for (int i=0;i<length;++i)
				chromosome.push_back(rightChromo[i]);
		}
	}

	// Chance of Mutation
	for (int i=0;i<length;++i)
	{
		if (RandFloat() <= MUTATION)
		{
			// Gaussian random perturbation utilizing Box-Mueller Transform
	 		float x1, x2, w, y1, y2;

 	    do {
 		   	x1 = 2.0 * ((rand()%10000)/10000.0) - 1.0;
 	     	x2 = 2.0 * ((rand()%10000)/10000.0) - 1.0;
 	     	w = x1 * x1 + x2 * x2;
   	 	} while ( w >= 1.0 );

   	 	w = sqrt( (-2.0 * log( w ) ) / w );
   	 	y1 = x1 * w;
    	y2 = x2 * w;

			chromosome[i] = chromosome[i] + sqrt(0.15) * y1;
		}
	}
 	Paddle<T> *offspring = new Paddle<T>(m_WorldID,m_SpaceID,m_StartPosition,true);		// constructor (world,space,x,y,z,in,hid,out,recur)
  offspring->setChromosome(chromosome);
	return offspring;
}

template <class T>
void Paddle<T>::enable(bool enabled)
{
}

template <class T>
void Paddle<T>::setStartPosition(dVector3 start)
{
	// call base function
	Organism<T>::setStartPosition(start);
  dReal x = start[0];
  dReal y = start[1];
  dReal z = start[2];

  // reset sensors and motors
  reset();

  // reset body positions
	dBodySetPosition (body[0],0+x,0+y,z);	// torso
	// right legs
	dBodySetPosition (body[1],x+TORSO[0]/2+LEG[0]/2,y+TORSO[1]/3,z-TORSO[2]/6);	//right forward leg
	// left legs
	dBodySetPosition (body[2],x-TORSO[0]/2-LEG[0]/2,y+TORSO[1]/3,z-TORSO[2]/6);	//left forward leg

	// reset rotations
  dMatrix3 R;
  dRFromAxisAndAngle (R,0,0,0,0);
  for (int i=0;i<numBody;++i)
		dBodySetRotation(body[i],R);

  // reset linear velocities
  for (int i=0;i<numBody;++i)
		dBodySetLinearVel(body[i],0,0,0);

  // reset angular velocities
  for (int i=0;i<numBody;++i)
		dBodySetAngularVel(body[i],0,0,0);
}

template <class T>
void Paddle<T>::constructBody(const dWorldID world, const dSpaceID space, const dVector3 position)
{
	// Body Setup
	TORSO[0] = 0.3; TORSO[1] = 0.3; TORSO[2] = 0.3;
	LEG[0] = 0.5; LEG[1] = 0.2; LEG[2] = 0.05; GAP = 0.01;

	l_foot_sens = false;					// init foot sensors
	r_foot_sens = false;
	body_sens = true;
	body = new dBodyID[numBody];	// allocate space for Bodies
	geom = new dGeomID[numGeom];	// allocate space for Geometries
	joint = new dJointID[numJoint];	// allocate space for Joints

	// create new Joint Group
	jGroup = dJointGroupCreate(0);
	//geomGroup = dCreateGeomGroup(space);

	// constants (Mass)
	const dReal MTORSO=2.8;
	const dReal MLEG=0.6;
	
  dReal x = position[0];
  dReal y = position[1];
  dReal z = position[2];
  
	dMass m;

	// create bodies
	for (int i=0;i<numBody;i++)
		body[i]=dBodyCreate(world);

	// following block
	dBodySetPosition (body[0],0+x,0+y,z);	// torso
	// right legs
	dBodySetPosition (body[1],x+TORSO[0]/2+LEG[0]/2,y+TORSO[1]/3,z-TORSO[2]/6);	//right forward leg
	// left legs
	dBodySetPosition (body[2],x-TORSO[0]/2-LEG[0]/2,y+TORSO[1]/3,z-TORSO[2]/6);	//left forward leg

	// torso
	dMassSetBox (&m,1,TORSO[0],TORSO[1],TORSO[2]);	// torso
	dMassAdjust (&m,MTORSO);	// torso weight
	dBodySetMass (body[0],&m);
	geom[0] = dCreateBox (space,TORSO[0],TORSO[1],TORSO[2]);
	dGeomSetBody (geom[0],body[0]);

	// legs
	dMassSetBox (&m,1,LEG[0],LEG[1],LEG[2]);	// legs
	dMassAdjust (&m,MLEG);	//  leg weight
	for (int i=1;i<=2;i++) {
		dBodySetMass (body[i],&m);
		geom[i] = dCreateBox (space,LEG[0],LEG[1],LEG[2]);
		dGeomSetBody (geom[i],body[i]);
	}


	/********************************** CREATE JOINTS **********************************/
	// Leg Sockets 1
	for (int i=0;i<=1;i++) {
		joint[i] = dJointCreateBall(world,jGroup);
		dJointAttach(joint[i],body[0],body[i+1]);
	}
	dJointSetBallAnchor(joint[0],x+TORSO[0]/2,y+TORSO[1]/3,z-TORSO[2]/4);
	dJointSetBallAnchor(joint[1],x-TORSO[0]/2,y+TORSO[1]/3,z-TORSO[2]/4);

	for (int i=2;i<=3;i++) {
		joint[i] = dJointCreateAMotor(world,jGroup);
		dJointAttach(joint[i],body[0],body[i-1]);
	}

	dJointSetAMotorParam(joint[2],dParamLoStop,-0.9); // back (right)
	dJointSetAMotorParam(joint[2],dParamHiStop,0.6);  // front (right)
	dJointSetAMotorParam(joint[3],dParamLoStop,-0.6); // front (left)
	dJointSetAMotorParam(joint[3],dParamHiStop,0.9);  // back (left)

	for (int i=2;i<=3;i++)
	{
		dJointSetAMotorMode(joint[i],dAMotorEuler);
		dJointSetAMotorAxis(joint[i],0,1,0,0,1);
		dJointSetAMotorAxis(joint[i],2,2,0,1,0);
		dJointSetAMotorParam(joint[i],dParamLoStop3,-0.5);  // up/
		dJointSetAMotorParam(joint[i],dParamHiStop3,0.5);   // /down
		dJointSetAMotorParam(joint[i],dParamVel, 0);
		dJointSetAMotorParam(joint[i],dParamFMax,0.8);
		dJointSetAMotorParam(joint[i],dParamLoStop2,-0.2);  //twist
		dJointSetAMotorParam(joint[i],dParamHiStop2,0.2);	// twist
		dJointSetAMotorParam(joint[i],dParamVel2,0);
		dJointSetAMotorParam(joint[i],dParamFMax2,0.8);
		dJointSetAMotorParam(joint[i],dParamVel3,0);
		dJointSetAMotorParam(joint[i],dParamFMax3,0.8);
		dJointSetAMotorParam(joint[i],dParamStopERP,0.3);
		dJointSetAMotorParam(joint[i],dParamStopCFM,0.05);
		dJointSetAMotorParam(joint[i],dParamStopERP2,0.3);
		dJointSetAMotorParam(joint[i],dParamStopCFM2,0.05);
		dJointSetAMotorParam(joint[i],dParamStopERP3,0.3);
		dJointSetAMotorParam(joint[i],dParamStopCFM3,0.05);
	}
}

template <class T>
void Paddle<T>::step(T milliseconds) const
{
//	SpikeNet<T> *temp = dynamic_cast<SpikeNet<T> *>(m_Brain);
//	temp->step(milliseconds);
}

template <class T>
void Paddle<T>::move()
{
	
}
  