/***************************************************************************
                          paddle.h  -  description
                             -------------------
    begin                : Wed Feb 18 2004
    copyright            : (C) 2004 by Gene Ruebsamen
    email                : gene@localhost.localdomain
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef PADDLE_H
#define PADDLE_H
#include <ode/ode.h>
#include <drawstuff/drawstuff.h>
#include "organism.h"
#include "../brains/neuralnet.h"
//#include "../brains/net2.h"
//#include "../brains/spike_net.h"
#include "../brains/brain.h"

template <class T>
class Paddle : public Organism<T> {
public:
	Paddle(dWorldID,dSpaceID,dVector3,bool=true);		// constructor (world,space,x,y,z,in,hid,out,recur)
	Paddle(const Paddle<T> &);
	~Paddle();	// destructor

	// cognitive functions
	vector<T> getSensoryData() const;
	inline vector<T> getBrainOutput(vector<T> &input);
	void activateMotors(const vector<T> &);
	void move();

	// GA Functions
	Organism<T> *operator+(const Organism<T> &right);		// reproduction
	const Paddle<T> &operator=(const Paddle<T> &right);		// assignment operator

	// Utility functions
	void groundContact(dGeomID);
	const dReal *getBodyRot() const;
	const double calcFitness() const;
	inline const dReal *getCurrentPosition() const;
	void setStartPosition(dVector3);
	void footContact(dGeomID,dGeomID,dGeomID);   		// are the feet touching the ground
	void enable(bool);
	void reset();
	void printStats() const;
	void draw(bool=false) const;
	inline void step(T ms) const;			// function for spiking neural network

private:
	void constructBody(const dWorldID,const dSpaceID,const dVector3);
	int numBody;
	int numGeom;
	int numJoint;
	bool l_foot_sens;
	bool r_foot_sens;
	bool body_sens;
	dBodyID *body;
	dGeomID *geom;
	dJointID *joint;
	dJointGroupID jGroup;
	dGeomID geomGroup;
	//  declarations
	dReal TORSO[3]; 	// length, width height
	dReal LEG[3];
	dReal GAP;
};

#endif

