/***************************************************************************
                          runner.cpp  -  description
                             -------------------
    begin                : Wed Jun 12 2002
    copyright            : (C) 2002 by Gene Ruebsamen
    email                : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "runner.h"      // import constants for bodies
#include <iostream>

using std::cout;
using std::endl;

Runner::Runner(dWorldID world, dSpaceID space, dReal x, dReal y, dReal z) :
	Creature(world,space,x,y,z), numBody(3), numGeom(3), numJoint(4)		// Build Creature
{
	// Body Setup
	TORSO[0] = 0.3; TORSO[1] = 0.3; TORSO[2] = 0.3;
	LEG[0] = 0.5; LEG[1] = 0.2; LEG[2] = 0.05; GAP = 0.01;

	l_foot_sens = false;					// init foot sensors
	r_foot_sens = false;
	body = new dBodyID[numBody];	// allocate space for Bodies
	geom = new dGeomID[numGeom];	// allocate space for Geometries	
	joint = new dJointID[numJoint];	// allocate space for Joints
	
	// create new Joint Group
	jGroup = dJointGroupCreate(0);
	//geomGroup = dCreateGeomGroup(space);

	// constants (Mass)
	const dReal MTORSO=2.8;
	const dReal MLEG=0.6;

	dMass m;

	// create bodies
	for (int i=0;i<numBody;i++)
		body[i]=dBodyCreate(world);

	// following block
	dBodySetPosition (body[0],0+x,0+y,z);	// torso
	// right legs
	dBodySetPosition (body[1],x+TORSO[0]/2+LEG[0]/2,y+TORSO[1]/3,z-TORSO[2]/6);	//right forward leg
	// left legs
	dBodySetPosition (body[2],x-TORSO[0]/2-LEG[0]/2,y+TORSO[1]/3,z-TORSO[2]/6);	//left forward leg
	
	// torso
	dMassSetBox (&m,1,TORSO[0],TORSO[1],TORSO[2]);	// torso
	dMassAdjust (&m,MTORSO);	// torso weight
	dBodySetMass (body[0],&m);
	geom[0] = dCreateBox (space,TORSO[0],TORSO[1],TORSO[2]);
	dGeomSetBody (geom[0],body[0]);

	// legs
	dMassSetBox (&m,1,LEG[0],LEG[1],LEG[2]);	// legs
	dMassAdjust (&m,MLEG);	//  leg weight
	for (int i=1;i<=2;i++) {
		dBodySetMass (body[i],&m);
		geom[i] = dCreateBox (space,LEG[0],LEG[1],LEG[2]);
		dGeomSetBody (geom[i],body[i]);
	}


	/********************************** CREATE JOINTS **********************************/
	// Leg Sockets 1
	for (int i=0;i<=1;i++) {
		joint[i] = dJointCreateBall(world,jGroup);
		dJointAttach(joint[i],body[0],body[i+1]);
	}
	dJointSetBallAnchor(joint[0],x+TORSO[0]/2,y+TORSO[1]/3,z-TORSO[2]/4);
	dJointSetBallAnchor(joint[1],x-TORSO[0]/2,y+TORSO[1]/3,z-TORSO[2]/4);

	for (int i=2;i<=3;i++) {
		joint[i] = dJointCreateAMotor(world,jGroup);
		dJointAttach(joint[i],body[0],body[i-1]);	
	}

	dJointSetAMotorParam(joint[2],dParamLoStop,-0.9); // back (right)
	dJointSetAMotorParam(joint[2],dParamHiStop,0.6);  // front (right)
	dJointSetAMotorParam(joint[3],dParamLoStop,-0.6); // front (left)
	dJointSetAMotorParam(joint[3],dParamHiStop,0.9);  // back (left)

	for (int i=2;i<=3;i++)
	{
		dJointSetAMotorMode(joint[i],dAMotorEuler);
		dJointSetAMotorAxis(joint[i],0,1,0,0,1);
		dJointSetAMotorAxis(joint[i],2,2,0,1,0);
		dJointSetAMotorParam(joint[i],dParamLoStop3,-0.5);  // up/
		dJointSetAMotorParam(joint[i],dParamHiStop3,0.5);   // /down
		dJointSetAMotorParam(joint[i],dParamVel, 0);
		dJointSetAMotorParam(joint[i],dParamFMax,0.8);
		dJointSetAMotorParam(joint[i],dParamLoStop2,-0.2);  //twist
		dJointSetAMotorParam(joint[i],dParamHiStop2,0.2);	// twist
		dJointSetAMotorParam(joint[i],dParamVel2,0);
		dJointSetAMotorParam(joint[i],dParamFMax2,0.8);
		dJointSetAMotorParam(joint[i],dParamVel3,0);
		dJointSetAMotorParam(joint[i],dParamFMax3,0.8);
		dJointSetAMotorParam(joint[i],dParamStopERP,0.3);
		dJointSetAMotorParam(joint[i],dParamStopCFM,0.05);
		dJointSetAMotorParam(joint[i],dParamStopERP2,0.3);
		dJointSetAMotorParam(joint[i],dParamStopCFM2,0.05);
		dJointSetAMotorParam(joint[i],dParamStopERP3,0.3);
		dJointSetAMotorParam(joint[i],dParamStopCFM3,0.05);
	}
}

Runner::~Runner()
{
	// destroy joint group
	dJointGroupDestroy(jGroup);
	for (int i=0;i<numGeom;i++) {
		dGeomDestroy(geom[i]);
	}
	// destroy bodies
	for (int i=0;i<numBody;i++)
		dBodyDestroy(body[i]);

	// deallocate space
	delete [] body;
	delete [] geom;
	delete [] joint;
}


void Runner::draw(bool selected) const
{
	dsSetColor (0,1,1);
	dsSetTexture (DS_WOOD); // DS_WOOD
	// head
	dsDrawBox(dBodyGetPosition(body[0]),dBodyGetRotation(body[0]),TORSO);
	if (selected)
		dsSetColor (1,0,0);
	else
		dsSetColor (1,0,1);
	// legs
	for (int i=1;i<=2;i++)	
		dsDrawBox(dBodyGetPosition(body[i]),dBodyGetRotation(body[i]),LEG);
}

void Runner::setMuscles(const float * const inputs)
{
	// left Arm Socket
	dJointSetAMotorParam(joint[2],dParamVel,(dReal) inputs[0]*1.5);
	dJointSetAMotorParam(joint[2],dParamFMax,(dReal)1);
	dJointSetAMotorParam(joint[2],dParamVel2,(dReal) inputs[1]);
	dJointSetAMotorParam(joint[2],dParamFMax2,(dReal)1);
	dJointSetAMotorParam(joint[2],dParamVel3,(dReal) inputs[2]*1.5);
	dJointSetAMotorParam(joint[2],dParamFMax3,(dReal)1);
	// right Arm socket
	dJointSetAMotorParam(joint[3],dParamVel,(dReal) inputs[3]*1.5);
	dJointSetAMotorParam(joint[3],dParamFMax,(dReal)1);
	dJointSetAMotorParam(joint[3],dParamVel2,(dReal) inputs[4]);
	dJointSetAMotorParam(joint[3],dParamFMax2,(dReal)1);
	dJointSetAMotorParam(joint[3],dParamVel3,(dReal) inputs[5]*1.5);
	dJointSetAMotorParam(joint[3],dParamFMax3,(dReal)1);
}

void Runner::getInputs(float * inputs) const
{
	// position inputs (position & rotation of head)
	const dReal * angle = dBodyGetRotation(body[0]);
	const dReal * vel = dBodyGetLinearVel(body[0]);
  const dReal * pos = dBodyGetPosition(body[0]);

	inputs[0] = (float)dJointGetAMotorAngle(joint[2],0);
	inputs[1] = (float)dJointGetAMotorAngle(joint[2],1);
	inputs[2] = (float)dJointGetAMotorAngle(joint[2],2);
	inputs[3] = (float)dJointGetAMotorAngle(joint[3],0);
	inputs[4] = (float)dJointGetAMotorAngle(joint[3],1);
	inputs[5] = (float)dJointGetAMotorAngle(joint[3],2);

/*	const dReal * bod = dBodyGetPosition(body[0]);
	const dReal * box = dBodyGetPosition(body[3]);

	// direction sensor
	float x = bod[0]-box[0];
	float y = bod[1]-box[1];
	float hyp = sqrt(pow(x,2)+pow(y,2));  */

	//inputs[6] = (float)asin(x/hyp);	// radians
	//inputs[7] = (float)hyp;
  	
	inputs[6]=angle[1];
	inputs[7]=vel[0];
//	inputs[0]=(float)fabs(vel[1]) + (float)fabs(vel[0]);
	// foot touch sensor
	if (r_foot_sens)
		inputs[8] = 1.0;
	else	
		inputs[8] = -1.0;
	if (l_foot_sens)
		inputs[9] = 1.0;
	else
		inputs[9] = -1.0;
	if (body_sens)
		inputs[10] = 1.0;
	else
		inputs[10] = -1.0;
}

void Runner::footContact(dGeomID o1, dGeomID o2, dGeomID ground)
{
	if (((o1==geom[1]) && (o2==ground)) || ((o2==geom[1])&&(o1==ground)))
		l_foot_sens=true;
	if (((o1==geom[2]) && (o2==ground)) || ((o2==geom[2])&&(o1==ground)))
		r_foot_sens=true;
}

void Runner::shutoff(void)
{
	dJointSetAMotorParam(joint[2],dParamVel,(dReal)0);
	dJointSetAMotorParam(joint[2],dParamFMax,(dReal)0);
	dJointSetAMotorParam(joint[3],dParamVel,(dReal)0);
	dJointSetAMotorParam(joint[3],dParamFMax,(dReal)0);	
}

void Runner::bodyContact(dGeomID o1, dGeomID o2, dGeomID ground)
{
	if (((o1==geom[0])&&(o2==ground)) || ((o2==geom[0])&&(o1==ground)))
		body_sens=true;
}

void Runner::clearSensor(void)
{
  r_foot_sens=false;
  l_foot_sens=false;
  body_sens=false;
}

const dReal *Runner::getHeadPos(void) const
{
	return dBodyGetPosition(body[0]);
}

const dReal *Runner::getBodyRot(void) const
{
	return dBodyGetRotation(body[0]);
}

const float Runner::getFitness(void) const
{
	const dReal *pos = getHeadPos();
//  float result;
/*  if (pos[1] > -3)
    result = -pos[1]*9 - pow(fabs(getStartX()-pos[0]),2);
  else
    result = 27 - ((-pos[1]-3)*9) - pow(fabs(getStartX()-pos[0]),2);
  */
	float result = (-pos[1]*9 - pow(fabs(getStartX()-pos[0]),4));

/*	const dReal * ideal = dBodyGetPosition(body[3]);

	float result = pow((10/(sqrt(pow((pos[0]-ideal[0]),2)+pow((pos[1]-ideal[1]),2)))/2),5); */
	if (result >= 0)
		return result;
	else
		return 0;		// y coordinate.
}

void Runner::randomForce(void) const
{
    dReal f[3];
    f[0] = 22.5 * (-0.5+1.0/RAND_MAX * rand());
    f[1] = 42.5 * (-0.5+1.0/RAND_MAX * rand());
    f[2] = .5 * (-0.5+1.0/RAND_MAX * rand());
    dBodyAddForce(body[0], f[0], f[1], f[2]);
}
void Runner::randomTorque(void) const
{
    dReal f[3];
    f[0] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
    f[1] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
    f[2] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
	  dBodyAddTorque(body[0], f[0], f[1], f[2]);
}

bool Runner::hasFallen(void) const
{
	return false;
}

void Runner::printStats(void) const
{
	const dReal *pos = dBodyGetPosition(body[0]);
	const dReal *rot = dBodyGetRotation(body[0]);
	const dReal * vel = dBodyGetLinearVel(body[0]);

	cout << "\nPosition:";
	cout << "X: " << pos[0] << " ";
	cout << "Y: " << pos[1] << " ";
	cout << "Z: " << pos[2] << endl;
	cout << "Rotation:" << endl;
	cout << "XR: " << rot[0] << " ";
	cout << "YR: " << rot[1] << " ";
	cout << "ZR: " << rot[2] << endl;
	cout << "Feet:" << endl;
	cout << "Right: " << r_foot_sens << " ";
	cout << "Left: " << l_foot_sens << " ";
  cout << "Body: " << body_sens << endl;
  cout << "Vel: 0:" << vel[0] << "  1:" << vel[1] << endl;
}
