/***************************************************************************
                          simplehuman.cpp  -  description
                             -------------------
    begin                : Mon Feb 18 2002
    copyright            : (C) 2002 by Gene Ruebsamen
    email                : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "simplehuman.h"      // import constants for bodies

SimpleHuman::SimpleHuman(dWorldID world, dSpaceID space, dReal x, dReal y, dReal z) :
	Creature(world,space,x,y,z), numBody(8), numGeom(4), numJoint(9)
{
	// setup
	HEAD = 0.12;         // x y z
	TORSO[0] = 0.4; TORSO[1] = 0.25; TORSO[2] = 0.50;
	LEG1[0] = 0.15; LEG1[1] = 0.15; LEG1[2] = 0.30;
	LEG2[0] = 0.15; LEG2[1] = 0.15; LEG2[2] = 0.25;
	FOOT[0] = 0.20; FOOT[1] = 0.40; FOOT[2] = 0.05;
	GAP = 0.01;

	l_foot_sens = false;					// init foot sensors
	r_foot_sens = false;
	body_sens = false;						// init body sensor
	fallen = false;
	body = new dBodyID[numBody];	// allocate space for Bodies
	geom = new dGeomID[numGeom];	// allocate space for Geometries	
	joint = new dJointID[numJoint];	// allocate space for Joints
	
	// create new Joint Group
	jGroup = dJointGroupCreate(0);
	// geomGroup = dCreateGeomGroup(space);

	// constants (Mass)
	const dReal MHEAD=0.1;
	const dReal MTORSO=1.3; //.7
	const dReal MULEG=0.6;  //.25
	const dReal MLLEG=0.4;
	const dReal MFOOT=0.1;

	dMass m;

	// create bodies
	for (int i=0;i<numBody;i++)
		body[i]=dBodyCreate(world);

	dBodySetPosition (body[0],0+x,0+y,z+TORSO[2]/2+HEAD);	// head
	dBodySetPosition (body[1],0+x,0+y,z);	// torso
	dBodySetPosition (body[2],TORSO[0]/3+x,0+y,z-(TORSO[2]/2+LEG1[2]/2+GAP));	//right upper leg
	dBodySetPosition (body[3],-TORSO[0]/3+x,0+y,z-(TORSO[2]/2+LEG1[2]/2+GAP));	// left upper leg
	dBodySetPosition (body[4],TORSO[0]/3+x,0+y,z-(TORSO[2]/2+LEG1[2]/2+LEG1[2]));	// right lower leg
	dBodySetPosition (body[5],-TORSO[0]/3+x,0+y,z-(TORSO[2]/2+LEG1[2]/2+LEG1[2]));	//left lower leg
	dBodySetPosition (body[6],TORSO[0]/3+GAP/2+x,-FOOT[0]/4+y,z-TORSO[2]/2-LEG1[2]-LEG2[2]-8*GAP);	//right foot
	dBodySetPosition (body[7],-TORSO[0]/3-GAP/2+x,-FOOT[0]/4+y,z-TORSO[2]/2-LEG1[2]-LEG2[2]-8*GAP);	//left foot

	// head
	dMassSetSphere(&m,1,HEAD);
	dMassAdjust(&m,MHEAD);
	dBodySetMass(body[0], &m);
	geom[0] = dCreateSphere(space,HEAD);
	dGeomSetBody(geom[0],body[0]);

	// torso
	dMassSetBox (&m,1,TORSO[0],TORSO[1],TORSO[2]);	// torso
	dMassAdjust (&m,MTORSO);	// torso weight
	dBodySetMass (body[1],&m);
	geom[1] = dCreateBox (space,TORSO[0],TORSO[1],TORSO[2]);
	dGeomSetBody (geom[1],body[1]);

	// upper legs
	dMassSetBox (&m,1,LEG1[0],LEG1[1],LEG1[2]);	// legs
	dMassAdjust (&m,MULEG);	// upper leg weight
	for (int i=2;i<=3;i++)
	{
		dBodySetMass (body[i],&m);
//		geom[i] = dCreateBox (space,LEG1[0],LEG1[1],LEG1[2]);
//		dGeomSetBody (geom[i],body[i]);
	}

	// lower legs
	dMassSetBox (&m,1,LEG2[0],LEG2[1],LEG2[2]);	// legs
  dMassAdjust (&m,MLLEG);	// lower leg weight
	for (int i=4;i<=5;i++)
	{
		dBodySetMass (body[i],&m);
//		geom[i] = dCreateBox (space,LEG2[0],LEG2[1],LEG2[2]);
//		dGeomSetBody (geom[i],body[i]);
	}

	// feet
	dMassSetBox (&m,1,FOOT[0],FOOT[1],FOOT[2]); 	// feet
	dMassAdjust (&m,MFOOT);	// feet weight
		geom[2] = dCreateBox (space,FOOT[0],FOOT[1],FOOT[2]);  // 0
		dGeomSetBody (geom[2],body[6]);
		geom[3] = dCreateBox (space,FOOT[0],FOOT[1],FOOT[2]);  // 0
		dGeomSetBody (geom[3],body[7]);

	for (int i=6;i<=7;i++)
	{
		dBodySetMass (body[i],&m);
//		geom[i] = dCreateBox (space,FOOT[0],FOOT[1],FOOT[2]);  // 0
//		dGeomSetBody (geom[i],body[i]);
	}

	// add geometries to GeomGroup
//	for (int i=0;i<numGeom;i++)
//		dGeomGroupAdd(geomGroup,geom[i]);


	/********************************** CREATE JOINTS **********************************/
	// head joint
	joint[0] = dJointCreateFixed(world,jGroup);		// add to joint group jGroup
	dJointAttach(joint[0],body[1],body[0]);
	dJointSetFixed(joint[0]);

	// right Leg Socket
	joint[1] = dJointCreateBall(world,jGroup);
	dJointAttach(joint[1],body[1],body[2]);
	dJointSetBallAnchor(joint[1],TORSO[0]/3+x,0+y,z-(TORSO[2]/2)-GAP/2);
	joint[7] = dJointCreateAMotor(world,jGroup);
	dJointAttach(joint[7],body[1],body[2]);

	// left Leg Socket
	joint[2] = dJointCreateBall(world,jGroup);
	dJointAttach(joint[2],body[1],body[3]);
	dJointSetBallAnchor(joint[2],-TORSO[0]/3+x,0+y,z-(TORSO[2]/2)-GAP/2);
	joint[8] = dJointCreateAMotor(world,jGroup);
	dJointAttach(joint[8],body[1],body[3]);

	dJointSetAMotorParam(joint[7],dParamLoStop3,-0.2); //left
	dJointSetAMotorParam(joint[7],dParamHiStop3,0.6);  // right
	dJointSetAMotorParam(joint[8],dParamLoStop3,-0.6); //left
	dJointSetAMotorParam(joint[8],dParamHiStop3,0.2);  // right

	for (int i=7;i<=8;i++)
	{
		dJointSetAMotorMode(joint[i],dAMotorEuler);
		dJointSetAMotorAxis(joint[i],0,1,0,0,1);
		dJointSetAMotorAxis(joint[i],2,2,0,1,0);
		dJointSetAMotorParam(joint[i],dParamLoStop,-0.2);  // twist
		dJointSetAMotorParam(joint[i],dParamHiStop,0.2);
		dJointSetAMotorParam(joint[i],dParamVel, 0);
		dJointSetAMotorParam(joint[i],dParamFMax,70);
		dJointSetAMotorParam(joint[i],dParamLoStop2,-0.3);  //back
		dJointSetAMotorParam(joint[i],dParamHiStop2,1.4);	// front
		dJointSetAMotorParam(joint[i],dParamVel2,0);
		dJointSetAMotorParam(joint[i],dParamFMax2,0.8);
		dJointSetAMotorParam(joint[i],dParamVel3,0);
		dJointSetAMotorParam(joint[i],dParamFMax3,0.8);
		dJointSetAMotorParam(joint[i],dParamStopERP,0.3);
		dJointSetAMotorParam(joint[i],dParamStopCFM,0.05);
		dJointSetAMotorParam(joint[i],dParamStopERP2,0.3);
		dJointSetAMotorParam(joint[i],dParamStopCFM2,0.05);
		dJointSetAMotorParam(joint[i],dParamStopERP3,0.3);
		dJointSetAMotorParam(joint[i],dParamStopCFM3,0.05);
	}
	// right knee
	joint[3] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[3],body[2],body[4]);
	dJointSetHingeAnchor(joint[3],TORSO[0]/3+x,y,z-(TORSO[2]/2+LEG1[2]));
	dJointSetHingeAxis(joint[3],1,0,0);
	dJointSetHingeParam(joint[3],dParamLoStop,-2.2);
	dJointSetHingeParam(joint[3],dParamHiStop,0);
	dJointSetHingeParam(joint[3],dParamVel,0);
	dJointSetHingeParam(joint[3],dParamFMax,0.5);
	dJointSetHingeParam(joint[3],dParamStopERP,0.4);
	dJointSetHingeParam(joint[3],dParamStopCFM,0.05);
	// left knee
	joint[4] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[4],body[3],body[5]);
	dJointSetHingeAnchor(joint[4],-TORSO[0]/3+x,y,z-(TORSO[2]/2+LEG1[2]));
	dJointSetHingeAxis(joint[4],1,0,0);
	dJointSetHingeParam(joint[4],dParamLoStop,-2.2);
	dJointSetHingeParam(joint[4],dParamHiStop,0);
	dJointSetHingeParam(joint[4],dParamVel,0);
	dJointSetHingeParam(joint[4],dParamFMax,0.5);
	dJointSetHingeParam(joint[4],dParamStopERP,0.4);
	dJointSetHingeParam(joint[4],dParamStopCFM,0.05);
	// right foot
	joint[5] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[5],body[4],body[6]);
	dJointSetHingeAnchor(joint[5],TORSO[0]/3+GAP/2+x,0.06+y,z-TORSO[2]/2-LEG1[2]-LEG2[2]-(8*GAP/2));
	dJointSetHingeAxis(joint[5],1,0,0);
	dJointSetHingeParam(joint[5],dParamLoStop,-0.8);
	dJointSetHingeParam(joint[5],dParamHiStop,0.3);
	dJointSetHingeParam(joint[5],dParamStopERP,0.3);
	dJointSetHingeParam(joint[5],dParamStopCFM,0.05);
	dJointSetHingeParam(joint[5],dParamBounce,0.3);
	dJointSetHingeParam(joint[5],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[5],dParamVel,0);
	dJointSetHingeParam(joint[5],dParamFMax,0.5);
	// left foot
	joint[6] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[6],body[5],body[7]);
	dJointSetHingeAnchor(joint[6],-TORSO[0]/3-GAP/2+x,0.06+y,z-TORSO[2]/2-LEG1[2]-LEG2[2]-(8*GAP/2));
	dJointSetHingeAxis(joint[6],1,0,0);
	dJointSetHingeParam(joint[6],dParamLoStop,-0.8);
	dJointSetHingeParam(joint[6],dParamHiStop,0.3);
	dJointSetHingeParam(joint[6],dParamStopERP,0.3);
	dJointSetHingeParam(joint[6],dParamStopCFM,0.05);
	dJointSetHingeParam(joint[6],dParamBounce,0.3);
	dJointSetHingeParam(joint[6],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[6],dParamVel,0);
	dJointSetHingeParam(joint[6],dParamFMax,0.5);
}

SimpleHuman::~SimpleHuman()
{
	// destroy joint group
	dJointGroupDestroy(jGroup);
	// destroy Geometries
	for (int i=0;i<numGeom;i++) {
//   	dGeomGroupRemove(geomGroup,geom[i]);
		dGeomDestroy(geom[i]);
	}
	// destroy bodies
	for (int i=0;i<numBody;i++)
		dBodyDestroy(body[i]);

	// deallocate space
	delete [] body;
	delete [] geom;
	delete [] joint;
}

void SimpleHuman::draw(bool selected) const
{
	dsSetColor (0,1,1);
	dsSetTexture (DS_WOOD); // DS_WOOD
	// head
	dsDrawSphere(dBodyGetPosition(body[0]),dBodyGetRotation(body[0]),HEAD);
	if (selected)
		dsSetColor (1,0,0);
	else
		dsSetColor (1,0,1);
	// torso
	dsDrawBox(dBodyGetPosition(body[1]),dBodyGetRotation(body[1]),TORSO);
	dsSetColor (1,1,0);
	// legs
	dsDrawBox(dBodyGetPosition(body[2]),dBodyGetRotation(body[2]),LEG1);
	dsDrawBox(dBodyGetPosition(body[3]),dBodyGetRotation(body[3]),LEG1);
	dsDrawBox(dBodyGetPosition(body[4]),dBodyGetRotation(body[4]),LEG2);
	dsDrawBox(dBodyGetPosition(body[5]),dBodyGetRotation(body[5]),LEG2);
	dsSetColor (0.5,0,1);
	// feet
	dsDrawBox(dBodyGetPosition(body[6]),dBodyGetRotation(body[6]),FOOT);
	dsDrawBox(dBodyGetPosition(body[7]),dBodyGetRotation(body[7]),FOOT);
}

void SimpleHuman::setMuscles(const float * const inputs)
{
	// right Leg Socket
	dJointSetAMotorParam(joint[7],dParamVel2,(dReal) inputs[0]*1.95);
	dJointSetAMotorParam(joint[7],dParamFMax2,(dReal)2.55);
	dJointSetAMotorParam(joint[7],dParamVel3,(dReal) inputs[1]*1.75);
	dJointSetAMotorParam(joint[7],dParamFMax3,(dReal)1.95);
	// left leg socket
	dJointSetAMotorParam(joint[8],dParamVel2,(dReal) inputs[2]*1.95);
	dJointSetAMotorParam(joint[8],dParamFMax2,(dReal)2.55);
	dJointSetAMotorParam(joint[8],dParamVel3,(dReal) inputs[3]*1.75);
	dJointSetAMotorParam(joint[8],dParamFMax3,(dReal)1.95);
	// right knee
	dJointSetHingeParam(joint[3],dParamVel,(dReal)inputs[4]*1.95);
	dJointSetHingeParam(joint[3],dParamFMax,(dReal)2.25);
	// left knee
	dJointSetHingeParam(joint[4],dParamVel,(dReal)inputs[5]*1.95);
	dJointSetHingeParam(joint[4],dParamFMax,(dReal)2.25);
	// right foot
	dJointSetHingeParam(joint[5],dParamVel,(dReal)inputs[6]);
	dJointSetHingeParam(joint[5],dParamFMax,(dReal)1.0);
	// left foot
	dJointSetHingeParam(joint[6],dParamVel,(dReal)inputs[7]);
	dJointSetHingeParam(joint[6],dParamFMax,(dReal)1.0);	
}

void SimpleHuman::getInputs(float * inputs) const
{
	// position inputs (position & rotation of head)
	const dReal * rot = dBodyGetAngularVel(body[6]);
//	const dReal * rot2 = dBodyGetRotation(body[7]);
	//const dReal * rot = dBodyGetLinearVel(body[0]);
	inputs[0] = (float)rot[0];
	inputs[1] = (float)rot[1];
	inputs[2] = (float)rot[2];
	// get right leg socket angles (3)
	inputs[3] = (float)dJointGetAMotorAngle(joint[7],0); // x-axis
	inputs[4] = (float)dJointGetAMotorAngle(joint[7],1); // y-axis
	inputs[5] = (float)dJointGetAMotorAngle(joint[7],2);
//	inputs[3] = (float)rot2[0];
//	inputs[4] = (float)rot2[1];
//	inputs[5] = (float)rot2[2];
	// get left leg socket angles (3)
	inputs[6] = (float)dJointGetAMotorAngle(joint[8],0);
	inputs[7] = (float)dJointGetAMotorAngle(joint[8],1);
	inputs[8] = (float)dJointGetAMotorAngle(joint[8],2);
	// get right knee angle
	inputs[9] = (float)dJointGetHingeAngleRate(joint[3]);
	// get left knee angle
	inputs[10] = (float)dJointGetHingeAngleRate(joint[4]);
	// get right foot angle
	inputs[11] = (float)dJointGetHingeAngleRate(joint[5]);
	// get left foot angle
	inputs[12] = (float)dJointGetHingeAngleRate(joint[6]);
	if (r_foot_sens) 	inputs[13] = 1.0;
	else	inputs[13] = -1.0;
	if (l_foot_sens)	inputs[14] = 1.0;
	else	inputs[14] = -1.0;
}

void SimpleHuman::footContact(dGeomID o1, dGeomID o2, dGeomID ground)
{
	if (((o1==geom[6]) && (o2==ground)) || ((o2==geom[6])&&(o1==ground)))
		r_foot_sens=true;
	else
		r_foot_sens=false;
	if (((o1==geom[7]) && (o2==ground)) || ((o2==geom[7])&&(o1==ground)))
		l_foot_sens=true;
	else
		l_foot_sens=false;
}

void SimpleHuman::bodyContact(dGeomID o1, dGeomID o2, dGeomID ground)
{
	if (((o1==geom[1]) && (o2==ground)) || ((o2==geom[1])&&(o1==ground)))
	{
		body_sens=true;
		fallen=true;
	}
	else if (((o1==geom[0]) && (o2==ground)) || ((o2==geom[0])&&(o1==ground)))
	{
		body_sens=true;
		fallen=true;
	}
	else
		body_sens=false;
}

void SimpleHuman::shutoff(void)
{
	// right Leg Socket
	dJointSetAMotorParam(joint[7],dParamVel2,(dReal)0);
	dJointSetAMotorParam(joint[7],dParamFMax2, (dReal)0);
	dJointSetAMotorParam(joint[7],dParamVel3,(dReal)0);
	dJointSetAMotorParam(joint[7],dParamFMax3, (dReal)0);
	// left leg socket
	dJointSetAMotorParam(joint[8],dParamVel2,(dReal)0);
	dJointSetAMotorParam(joint[8],dParamFMax2,(dReal)0);
	dJointSetAMotorParam(joint[8],dParamVel3,(dReal)0);
	dJointSetAMotorParam(joint[8],dParamFMax3,(dReal)0);
	// right knee
	dJointSetHingeParam(joint[3],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[3],dParamFMax,(dReal)0);
	// left knee
	dJointSetHingeParam(joint[4],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[4],dParamFMax,(dReal)0);	
	// right foot
	dJointSetHingeParam(joint[5],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[5],dParamFMax,(dReal)0);
	// left foot
	dJointSetHingeParam(joint[6],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[6],dParamFMax,(dReal)0);	

}

float SimpleHuman::getFitness(void) const
{

	const dReal * pos = getHeadPos();
	float fitness;

	// fitness function
	if (fabs(pos[1]) > 10 || (fabs(pos[2]) > 5) || (fabs(pos[0]-getStartX()) > 6))
		fitness = 0;
	else if (hasFallen())
		fitness = -pos[1]+pos[2]*8;  	// if fallen down
	else
		fitness = -pos[1]*5+pos[2]*8;	// if still standing

	return fitness;
}

const dReal *SimpleHuman::getHeadPos(void) const
{
	return dBodyGetPosition(body[0]);
}

const dReal *SimpleHuman::getRFootPos(void) const
{
	return dBodyGetPosition(body[6]);
}

const dReal *SimpleHuman::getLFootPos(void) const
{
	return dBodyGetPosition(body[7]);
}

bool SimpleHuman::hasFallen(void) const
{
	return fallen;
}

void SimpleHuman::clearSensor(void)
{
  r_foot_sens=false;
  l_foot_sens=false;
  body_sens=false;
}
