/***************************************************************************
                          world.cpp  -  description
                             -------------------
    begin                : Mon Feb 9 2004
    copyright            : (C) 2004 by Gene Ruebsamen
    email                : gene@localhost.localdomain
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "world.h"

World::World(dVector3 gravity)
{
  m_WorldID = dWorldCreate();

  //dWorldSetGravity(m_WorldID, gravity);
  dWorldSetERP(m_WorldID,ERP);
  dWorldSetCFM(m_WorldID,CFM);
  m_CollisionJointGroup = dJointGroupCreate(0);
  // create top level space
  m_SpaceID = dHashSpaceCreate(0);
    
//  return m_WorldID;
}

World::~World()
{
	dSpaceDestroy(m_SpaceID);
	dJointGroupDestroy(m_CollisionJointGroup);
	dWorldDestroy(m_WorldID);
}

dWorldID World::getWorldID()
{
  return m_WorldID;
}

dSpaceID World::getSpaceID()
{
	return m_SpaceID;
}

// Remember, this function is static, it has no "this" pointer,
// nor any of the other members of the World class.
//
void World::CollisionCallback (void *pData, dGeomID o1, dGeomID o2)
{
	// pData is the "this" that was passed into dSpaceCollide
	World* pThis = (World*) pData;
	if (pThis)
		pThis->CollisionCallback (o1, o2);
}

// This is a nonstatic member function, here you have access to
// all of the member variables in your World class.  Notice the
// parameters to dJointCreateContact, near the bottom.
//
void World::CollisionCallback (dGeomID o1, dGeomID o2)
{
	// return if objects are null
	if (!(o1 && o2))
	{
		//Breakpoint ("CollisionCallback with null geometry\n");
		return;
	}

	// handle space to space or space to geom collisions
	if (dGeomIsSpace (o1) || dGeomIsSpace (o2))
	{
		// colliding a space with something
		dSpaceCollide2 (o1,o2,this,CollisionCallback);
		// collide all geoms internal to the space(s)
		if (dGeomIsSpace (o1)) dSpaceCollide ((dSpaceID)o1,this,CollisionCallback);
		if (dGeomIsSpace (o2)) dSpaceCollide ((dSpaceID)o2,this,CollisionCallback);
	}
	else // handle geom collisions (non-space)
	{
		// ignore collisions between bodies that are connected by the same joint
		dBodyID Body1 = NULL, Body2 = NULL;

		if (o1) Body1 = dGeomGetBody (o1);
		if (o2) Body2 = dGeomGetBody (o2);

		if (Body1 && Body2 && dAreConnected (Body1, Body2))
			return;

		// colliding two non-space geoms, so generate contact
		// points between o1 and o2
		const int iContacts = ODE_MAX_CONTACTS;
		dContact contact[iContacts];

		int iCount = dCollide (o1, o2, iContacts, &contact[0].geom, sizeof (dContact));

		if (iCount)
		{
			int iMax = min(iCount, iContacts);
			for (int iContact = 0; iContact < iMax; iContact++)
			{
/*				contact[iContact].surface.mode = dContactSoftERP | dContactSoftCFM | dContactApprox1;//dContactSlip1 | dContactSlip2;
				contact[iContact].surface.mu = 0.5;
				//contact[iContact].surface.slip1 = 0.2; // lower is more friction 0.8
				//contact[iContact].surface.slip2 = 0.2;
				contact[iContact].surface.soft_erp = 0.2;
				contact[iContact].surface.soft_cfm = 0.05;
  */

				contact[iContact].surface.mode = dContactSoftERP | dContactSoftCFM | dContactSlip1 | dContactSlip2;
				contact[iContact].surface.mu = dInfinity;
				contact[iContact].surface.slip1 = 0.2; // lower is more friction 0.8
				contact[iContact].surface.slip2 = 0.2;
				contact[iContact].surface.soft_erp = 0.2;
				contact[iContact].surface.soft_cfm = 0.05;

				dJointID ContactJoint = dJointCreateContact (m_WorldID, m_CollisionJointGroup, &contact[iContact]);

				dJointAttach (ContactJoint, Body1, Body2);
			}
		}
	}
}

// Step the world.  Notice that "this" is passed as the second
// parameter to dSpaceCollide.  That's crucial.
//
bool World::fStep (unsigned int uMilliseconds)
{
	// check for collisions, create contact joints
	dSpaceCollide (m_SpaceID, this, CollisionCallback);

	// convert to seconds
	dReal rElapsedTime = ((dReal) uMilliseconds) / ((dReal) 1000.0);

	// step the world
	dWorldStep (m_WorldID, rElapsedTime);

	// remove all contact joints
	dJointGroupEmpty (m_CollisionJointGroup);

  return true;//	return fReturn;
}


// Step the world.  Notice that "this" is passed as the second
// parameter to dSpaceCollide.  That's crucial.
//
bool World::fStepFast (unsigned int uMilliseconds, unsigned int iterations)
{
	// check for collisions, create contact joints
	dSpaceCollide (m_SpaceID, this, CollisionCallback);

	// convert to seconds
	dReal rElapsedTime = ((dReal) uMilliseconds) / ((dReal) 1000.0);

	// step the world
	dWorldStepFast1 (m_WorldID, rElapsedTime, iterations);

	// remove all contact joints
	dJointGroupEmpty (m_CollisionJointGroup);

	return true; //return fReturn;
}
